finnhub_exchanges_url: str = "https://raw.githubusercontent.com/amor71/LiuAlgoTrader/4520081c56aef7a23b37416ae6e95465b5e50f80/static/finnhub_exchanges.csv"
finnhub_indices_url: str = "https://raw.githubusercontent.com/amor71/LiuAlgoTrader/4520081c56aef7a23b37416ae6e95465b5e50f80/static/finnhub_indices.csv"



import sys
import traceback
from datetime import date, timedelta
from typing import Dict

from liualgotrader.analytics import consolidate
from liualgotrader.common.tlog import tlog
from liualgotrader.miners.base import Miner
from liualgotrader.models.algo_run import AlgoRun


class Gainloss(Miner):
    def __init__(
        self,
        data: Dict,
        debug=False,
    ):
        try:
            self.days = int(data["days"])
            self.debug = debug
        except Exception:
            raise ValueError(
                "[ERROR] Miner must receive positive `days` parameter"
            )
        super().__init__(name="GainLossMiner")

    async def run(self) -> bool:
        data = await AlgoRun.get_batch_ids(
            start_date=date.today() - timedelta(days=self.days)
        )
        tlog(f"Miner {self.name} will consolidate {len(data)} batches")
        for index, row in data.iterrows():
            tlog(f"{int(index)+1}/{len(data)}")
            try:
                await consolidate.trades(row.batch_id)
            except Exception as e:
                tlog(f"[ERROR] aborted w/ exception {e}")
                exc_info = sys.exc_info()
                traceback.print_exception(*exc_info)
                del exc_info
                raise

        return True



from datetime import date, datetime
from typing import Dict, List, Set

import numpy as np
import pandas as pd
import pandas_market_calendars
import pytz
import requests
from pandas import DataFrame as df

from liualgotrader.common import config
from liualgotrader.common.data_loader import DataLoader  # type: ignore
from liualgotrader.common.data_loader import m_and_a_data  # type: ignore
from liualgotrader.common.tlog import tlog
from liualgotrader.common.types import TimeScale
from liualgotrader.fincalcs.vwap import add_daily_vwap

volume_today: Dict[str, int] = {}
quotes: Dict[str, df] = {}

NY = "America/New_York"
nytz = pytz.timezone(NY)


def get_symbol_data(
    symbol: str,
    start_date: date,
    end_date: date,
    data_loader=None,
    scale="minute",
) -> df:
    if not data_loader:
        data_loader = DataLoader(TimeScale[scale])

    return data_loader[symbol][start_date:end_date]  # type: ignore


async def get_sector_tickers(sector: str) -> List[str]:
    async with config.db_conn_pool.acquire() as conn:
        async with conn.transaction():
            records = await conn.fetch(
                """
                    SELECT symbol
                    FROM ticker_data
                    WHERE sector = $1
                """,
                sector,
            )

            return [record[0] for record in records if record[0]]


async def get_sectors_tickers(sectors: List[str]) -> List[str]:
    async with config.db_conn_pool.acquire() as conn:
        async with conn.transaction():
            q = f"""
                    SELECT symbol
                    FROM ticker_data
                    WHERE sector in ({str(sectors)[1:-1]})
                """

            records = await conn.fetch(q)

    return [record[0] for record in records if record[0]]


async def get_industry_tickers(industry: str) -> List[str]:
    async with config.db_conn_pool.acquire() as conn:
        records = await conn.fetch(
            """
                SELECT symbol
                FROM ticker_data
                WHERE industry = $1
            """,
            industry,
        )

    return [record[0] for record in records if record[0]]


async def get_industries_tickers(industries: List[str]) -> List[str]:
    async with config.db_conn_pool.acquire() as conn:
        q = f"""
                SELECT symbol
                FROM ticker_data
                WHERE industry in ({str(industries)[1:-1]})
            """

        records = await conn.fetch(q)

    return [record[0] for record in records if record[0]]


async def get_market_sectors() -> List[str]:
    async with config.db_conn_pool.acquire() as conn:
        records = await conn.fetch(
            """
                SELECT DISTINCT sector
                FROM ticker_data
            """
        )
        return [record[0] for record in records if record[0]]


async def get_market_industries() -> List[str]:
    async with config.db_conn_pool.acquire() as conn:
        records = await conn.fetch(
            """
                SELECT DISTINCT industry
                FROM ticker_data
            """
        )

        return [record[0] for record in records if record[0]]


async def sp500_historical_constituents(date: str):
    tlog(f"loading sp500 constituents for {date}")
    table = pd.read_html(
        "https://en.wikipedia.org/wiki/List_of_S%26P_500_companies"
    )
    symbols: List = table[0].Symbol.to_list()
    changes = table[1]

    changes["date"] = changes.Date.apply(
        lambda x: datetime.strptime(x[0], "%B %d, %Y"), axis=1
    )
    m_and_a_data.index = m_and_a_data.index.astype("datetime64[ns]", copy=True)

    adjusted_symbols = m_and_a_data.loc[date < m_and_a_data.index, "to_symbol"]
    print("adjusted_symbols", adjusted_symbols)
    changes = changes.loc[changes.date > date]

    unadusted: Set = set(symbols)
    while True:
        no_changes = True
        for _, row in changes.iterrows():
            if not row.Added.dropna().empty and row.Added.Ticker in unadusted:
                unadusted.remove(row.Added.Ticker)
                no_changes = False
            if (
                not row.Removed.dropna().empty
                and row.Removed.Ticker in unadusted
            ):
                unadusted.add(row.Removed.Ticker)
                no_changes = False

        if no_changes:
            break

    while True:
        no_changes = True
        for symbol in adjusted_symbols:
            if symbol in unadusted:
                unadusted.add(
                    m_and_a_data.loc[
                        m_and_a_data.to_symbol == symbol, "from_symbol"
                    ].item()
                )
                unadusted.remove(symbol)
                no_changes = False

        if no_changes:
            break

    return list(unadusted)


async def get_trading_holidays() -> List[str]:
    nyse = pandas_market_calendars.get_calendar("NYSE")
    return nyse.holidays().holidays


async def get_trading_day(now: date, offset: int) -> date:
    cbd_offset = pd.tseries.offsets.CustomBusinessDay(
        n=-offset, holidays=await get_trading_holidays()
    )

    return (now + cbd_offset).date()



"""
Get Market data from Data Providers and pump to consumers
"""
import asyncio
import inspect
import json
import os
import random
import sys
import traceback
from datetime import datetime
from multiprocessing import Queue
from queue import Empty
from typing import Dict, List

from liualgotrader.common import config
from liualgotrader.common.database import create_db_connection
from liualgotrader.common.tlog import tlog, tlog_exception
from liualgotrader.common.types import QueueMapper, WSEventType
from liualgotrader.data.data_factory import streaming_factory
from liualgotrader.models.tradeplan import TradePlan
from liualgotrader.models.trending_tickers import TrendingTickers
from liualgotrader.trading.trader_factory import trader_factory

last_msg_tstamp: datetime = datetime.now()
symbols: List[str] = []
queue_id_hash: Dict[str, int] = {}
symbol_strategy: Dict = {}


def get_new_symbols_and_queues(
    symbols_details: Dict, queues: List[Queue], num_consumer_processes: int
) -> List[str]:
    global symbol_strategy
    global symbols

    new_symbols: List = []
    for symbol_details in symbols_details:
        symbol = symbol_details["symbol"].lower()
        if symbol not in symbols:
            new_symbols.append(symbol)
            symbol_strategy[symbol] = symbol_details["target_strategy_name"]
            streaming_factory().get_instance().queues[symbol] = queues[
                random.SystemRandom().randint(0, num_consumer_processes - 1)
            ]

    return new_symbols


async def subscribe_new_symbols(new_symbols: List[str]):
    await streaming_factory().get_instance().subscribe(
        new_symbols,
        [
            WSEventType.SEC_AGG,
            WSEventType.MIN_AGG,
            WSEventType.TRADE,
        ],
    )


async def scanners_iteration(
    scanner_queue: Queue,
    queues: List[Queue],
    num_consumer_processes: int,
):
    global symbols
    symbols_details = scanner_queue.get(timeout=1)
    if len(
        new_symbols := get_new_symbols_and_queues(
            symbols_details=json.loads(symbols_details),
            queues=queues,
            num_consumer_processes=num_consumer_processes,
        )
    ):
        await subscribe_new_symbols(new_symbols)
        trending_db = TrendingTickers(config.batch_id)
        await trending_db.save(new_symbols)
        symbols += new_symbols
        tlog(
            f"added {len(new_symbols)}:{new_symbols[:20]}..{new_symbols[-20:]} TOTAL: {len(symbols)}"
        )
        await asyncio.sleep(1)


async def scanner_input(
    scanner_queue: Queue,
    queues: List[Queue],
    num_consumer_processes: int,
) -> None:
    tlog("scanner_input() task starting ")

    while True:
        try:
            await scanners_iteration(
                scanner_queue=scanner_queue,
                queues=queues,
                num_consumer_processes=num_consumer_processes,
            )
        except Empty:
            await asyncio.sleep(30)
        except asyncio.CancelledError:
            tlog("scanner_input() task task cancelled ")
            break
        except Exception as e:
            tlog(
                f"Exception in scanner_input(): exception of type {type(e).__name__} with args {e.args}"
            )
            if config.debug_enabled:
                exc_info = sys.exc_info()
                lines = traceback.format_exception(*exc_info)
                for line in lines:
                    tlog(f"error: {line}")
                traceback.print_exception(*exc_info)
                del exc_info

    tlog("scanner_input() task completed")


async def trade_run(qm: QueueMapper) -> None:
    at = trader_factory(qm)
    tlog(f"trade_run() starting using {at} trading")
    await at.run()
    tlog("trade_run() completed")


async def dispense_strategies(
    consumer_queues: List[Queue], tradeplan_entries: List[TradePlan]
) -> None:
    for tradeplan in tradeplan_entries:
        happy_consumer = consumer_queues[
            random.SystemRandom().randint(0, len(consumer_queues) - 1)
        ]

        payload = {
            "EV": "new_strategy",
            "parameters": tradeplan.parameters,
            "portfolio_id": tradeplan.portfolio_id,
        }

        try:
            happy_consumer.put(payload)
        except Exception as e:
            tlog(
                f"Exception in dispense_strategies: exception of type {type(e).__name__} with args {e.args}"
            )
            if config.debug_enabled:
                tlog_exception(str(e))


async def tradeplan_scanner(
    consumer_queues: List[Queue],
) -> None:
    tlog("tradeplan_scanner() task starting ")

    last_scan: datetime = datetime.utcnow()
    while True:
        await asyncio.sleep(5 * 60)

        tlog(
            f"tradeplan_scanner(): check for new execution requests since {last_scan}"
        )
        try:
            new_tradeplan_entries = await TradePlan.get_new_entries(
                since=last_scan
            )
            last_scan = datetime.utcnow()

            if new_tradeplan_entries:
                tlog(
                    f"found {len(new_tradeplan_entries)} new tradeplan entries for execution"
                )
                await dispense_strategies(
                    consumer_queues, new_tradeplan_entries
                )
        except Exception as e:
            tlog(f"[EXCEPTION] tradeplan_scanner() {e}")
            if config.debug_enabled:
                tlog_exception(str(e))

    tlog("tradeplan_scanner() task completed")


async def run(
    queues: List[Queue],
    qm: QueueMapper,
) -> None:
    global queue_id_hash

    ps = streaming_factory()(qm)
    await ps.run()
    for symbol in symbols:
        symbol = symbol.lower()
        qm[symbol] = queues[queue_id_hash[symbol]]
    await ps.subscribe(
        symbols, [WSEventType.SEC_AGG, WSEventType.MIN_AGG, WSEventType.TRADE]
    )


"""
process main
"""


async def producer_async_main(
    queues: List[Queue],
    scanner_queue: Queue,
    num_consumer_processes: int,
):
    await create_db_connection(str(config.dsn))
    qm = QueueMapper(queue_list=queues)
    await run(queues=queues, qm=qm)

    # TODO: Support multiple brokers
    at = trader_factory(qm)
    trade_updates_task = await at.run()

    scanner_input_task = asyncio.create_task(
        scanner_input(scanner_queue, queues, num_consumer_processes),
        name="scanner_input",
    )

    tradeplan_scanner_task = asyncio.create_task(
        tradeplan_scanner(queues), name="tradeplan_scanner"
    )

    async_tasks_to_gather = [scanner_input_task, tradeplan_scanner_task]

    if inspect.iscoroutinefunction(trade_updates_task):
        async_tasks_to_gather.append(trade_updates_task)  # type: ignore

    await asyncio.gather(
        *async_tasks_to_gather,
        return_exceptions=True,
    )

    tlog("producer_async_main() completed")


def producer_main(
    unique_id: str,
    queues: List[Queue],
    conf_dict: Dict,
    scanner_queue: Queue,
    num_consumer_processes: int,
) -> None:
    tlog(f"*** producer_main() starting w pid {os.getpid()} ***")
    try:
        config.batch_id = unique_id
        asyncio.run(
            producer_async_main(queues, scanner_queue, num_consumer_processes)
        )
    except KeyboardInterrupt:
        tlog("producer_main() - Caught KeyboardInterrupt")
    except Exception as e:
        tlog(
            f"producer_main() - exception of type {type(e).__name__} with args {e.args}"
        )
        if config.debug_enabled:
            traceback.print_exc()

    tlog("*** producer_main() completed ***")



import json
from typing import Dict, List, Optional, Tuple

import tabulate
from pandas import DataFrame

from liualgotrader.common import config
from liualgotrader.common.database import create_db_connection
from liualgotrader.common.types import AssetType
from liualgotrader.models.accounts import Accounts


class Portfolio:
    def __init__(
        self,
        portfolio_id: str,
        portfolio_size: float,
        asset_type: AssetType,
        account_id: Optional[int],
        parameters: Dict,
    ):
        self.portfolio_id = portfolio_id
        self.portfolio_size = portfolio_size
        self.parameters = parameters
        self.account_id = account_id
        self.asset_type = AssetType[str(asset_type)]

    def __str__(self):
        return f"id={self.portfolio_id} account-id={self.account_id} size={self.portfolio_size} asset_type={self.asset_type}, params={self.parameters}"

    @classmethod
    async def load_by_batch_id(cls, batch_id: str):
        try:
            pool = config.db_conn_pool
        except AttributeError:
            await create_db_connection()
            pool = config.db_conn_pool

        async with pool.acquire() as con:
            data = await con.fetchrow(
                """
                    SELECT p.portfolio_id, p.size, p.assets, p.account_id, p.parameters
                    FROM 
                        portfolio as p, portfolio_batch_ids as b
                    WHERE
                        p.portfolio_id = b.portfolio_id
                        AND b.batch_id = $1
                """,
                batch_id,
            )

            if data:
                return Portfolio(*data)

    @classmethod
    async def load_by_portfolio_id(cls, portfolio_id: str):
        try:
            pool = config.db_conn_pool
        except AttributeError:
            await create_db_connection()
            pool = config.db_conn_pool

        async with pool.acquire() as con:
            data = await con.fetchrow(
                """
                    SELECT p.portfolio_id, p.size, p.assets, p.account_id, p.parameters
                    FROM 
                        portfolio as p
                    WHERE
                        p.portfolio_id = $1
                """,
                portfolio_id,
            )

            if data:
                return Portfolio(*data)

    @classmethod
    async def save(
        cls,
        portfolio_id: str,
        portfolio_size: float,
        credit: float,
        parameters: Dict,
        asset_type: AssetType = AssetType.US_EQUITIES,
        external_account_id: Optional[str] = None,
        broker: Optional[str] = None,
    ):
        if not hasattr(config, "db_conn_pool"):
            await create_db_connection()
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            async with con.transaction():
                account_id = await Accounts.create(
                    portfolio_size,
                    allow_negative=credit > 0.0,
                    credit_line=credit,
                    details=parameters,
                )
                await con.execute(
                    """
                        INSERT INTO portfolio (portfolio_id, size, account_id, assets, parameters, external_account_id, broker)
                        VALUES ($1, $2, $3, $4, $5, $6, $7)
                    """,
                    portfolio_id,
                    portfolio_size,
                    account_id,
                    asset_type.name,
                    json.dumps(parameters),
                    external_account_id,
                    broker,
                )

    @classmethod
    async def associate_batch_id_to_profile(
        cls, portfolio_id: str, batch_id: str
    ) -> None:
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            await con.execute(
                """
                    INSERT INTO portfolio_batch_ids (portfolio_id, batch_id)
                    VALUES ($1, $2)
                """,
                portfolio_id,
                batch_id,
            )

    @classmethod
    async def is_associated(
        cls, portfolio_id: str, batch_id: Optional[str] = None
    ) -> bool:
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            return (
                await con.fetchval(
                    """
                        SELECT EXISTS (
                            SELECT 1 
                            FROM portfolio_batch_ids
                            WHERE portfolio_id = $1 AND batch_id=$2
                        )
                    """,
                    portfolio_id,
                    batch_id,
                )
                if batch_id
                else await con.fetchval(
                    """
                        SELECT EXISTS (
                            SELECT 1 
                            FROM portfolio_batch_ids
                            WHERE portfolio_id = $1
                        )
                    """,
                    portfolio_id,
                )
            )

    @classmethod
    async def exists(cls, portfolio_id: str) -> bool:
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            result = await con.fetchval(
                """
                    SELECT EXISTS (
                        SELECT 1 
                        FROM portfolio
                        WHERE portfolio_id = $1
                    )
                """,
                portfolio_id,
            )
            return result

    @classmethod
    async def get_portfolio_account_balance(cls, portfolio_id: str) -> float:
        pool = config.db_conn_pool

        async with pool.acquire() as con:
            return await con.fetchval(
                """
                    SELECT
                        balance
                    FROM 
                        accounts as a, portfolio as f 
                    WHERE
                        a.account_id = f.account_id
                        AND portfolio_id = $1
                """,
                portfolio_id,
            )

    @classmethod
    async def get_external_account_id(
        cls, portfolio_id: str
    ) -> Tuple[Optional[str], Optional[str]]:
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            r = await con.fetchrow(
                """
                    SELECT
                        external_account_id, broker
                    FROM 
                        portfolio
                    WHERE
                        portfolio_id = $1
                """,
                portfolio_id,
            )

            return (r["external_account_id"], r["broker"])

    @classmethod
    async def list_portfolios(cls) -> List[str]:
        try:
            pool = config.db_conn_pool
        except AttributeError:
            await create_db_connection()
            pool = config.db_conn_pool

        async with pool.acquire() as con:
            async with con.transaction():
                rows = await con.fetch(
                    """
                        SELECT
                            portfolio_id , a.tstamp as last_transaction, size, parameters ,assets, p.tstamp 
                        FROM
                            portfolio as p
                        LEFT JOIN account_transactions as a
                        ON
                            p.account_id = a.account_id
                        WHERE
                            p.expire_tstamp is NULL
                        ORDER BY 
                            p.tstamp DESC
                    """,
                )
                return DataFrame(
                    data=rows,
                    columns=[
                        "portfolio_id",
                        "last_transaction",
                        "size",
                        "parameters",
                        "assets",
                        "tstamp",
                    ],
                )



import datetime
import json
from typing import Dict

import asyncpg
import pandas as pd

from liualgotrader.common import config
from liualgotrader.common.database import fetch_as_dataframe
from liualgotrader.common.tlog import tlog


class Accounts:
    @classmethod
    async def create(
        cls,
        balance: float,
        allow_negative: bool = False,
        credit_line: float = 0.0,
        details: Dict = {},
    ) -> int:
        q = """
                INSERT INTO 
                    accounts (balance, allow_negative, credit_line, details)
                VALUES
                    ($1, $2, $3, $4)
                RETURNING account_id;
            """
        params = [balance, allow_negative, credit_line, json.dumps(details)]
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            async with con.transaction():
                result = await con.fetchval(q, *params)

        return result

    @classmethod
    async def check_if_enough_balance_to_withdraw(
        cls, account_id: int, potential_withdraw: float
    ) -> bool:
        if potential_withdraw < 0:
            raise AssertionError(
                f"check_if_enough_balance(): potential transaction amount ({potential_withdraw}) can't be negative "
            )

        query = """
            SELECT
                balance, allow_negative, credit_line
            FROM 
                accounts 
            WHERE
                account_id = $1
        """
        rc = False
        async with config.db_conn_pool.acquire() as con:
            row = await con.fetchrow(query, account_id)
            if row["balance"] > potential_withdraw:
                rc = True
            elif (
                row["allow_negative"]
                and row["balance"] + row["credit_line"] > potential_withdraw
            ):
                rc = True

        return rc

    @classmethod
    async def get_balance(cls, account_id: int) -> float:
        pool = config.db_conn_pool

        async with pool.acquire() as con:
            result = await con.fetchval(
                """
                    SELECT
                        balance
                    FROM 
                        accounts 
                    WHERE
                        account_id = $1
                """,
                account_id,
            )
        return result

    @classmethod
    async def add_transaction(
        cls,
        account_id: int,
        amount: float,
        tstamp: datetime.datetime = None,
        details: Dict = {},
    ):
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            async with con.transaction():
                try:
                    if tstamp:
                        await con.execute(
                            """
                                INSERT INTO 
                                    account_transactions (account_id, amount, tstamp, details)
                                VALUES
                                    ($1, $2, $3, $4);
                            """,
                            account_id,
                            amount,
                            tstamp,
                            json.dumps(details),
                        )
                    else:
                        await con.execute(
                            """
                                INSERT INTO 
                                    account_transactions (account_id, amount, details)
                                VALUES
                                    ($1, $2, $3);
                            """,
                            account_id,
                            amount,
                            json.dumps(details),
                        )
                except Exception as e:
                    tlog(
                        f"[EXCEPTION] add_transaction({account_id}, {amount}, {tstamp} ) failed with {e} current balance={await cls.get_balance(account_id)}"
                    )
                    raise

    @classmethod
    async def get_transactions(cls, account_id: int) -> pd.DataFrame:
        q = """
            SELECT
                tstamp, amount
            FROM 
                account_transactions 
            WHERE
                account_id = $1
            """

        df = await fetch_as_dataframe(q, account_id)
        return (
            df.set_index("tstamp", drop=True).sort_index()
            if not df.empty
            else df
        )

    @classmethod
    async def clear_balance(cls, account_id: int, new_balance: float):
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            async with con.transaction():
                try:
                    await con.execute(
                        """
                            UPDATE  
                                accounts
                            SET
                                balance = $2
                            WHERE
                                account_id = $1
                        """,
                        account_id,
                        new_balance,
                    )
                except Exception as e:
                    tlog(
                        f"[EXCEPTION] clear_balance() failed with {e} current balance={await cls.get_balance(account_id)}"
                    )
                    raise

    @classmethod
    async def clear_account_transactions(cls, account_id: int):
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            async with con.transaction():
                try:
                    await con.execute(
                        """
                            DELETE FROM  
                                account_transactions
                            WHERE
                                account_id = $1
                        """,
                        account_id,
                    )
                except Exception as e:
                    tlog(
                        f"[EXCEPTION] clear_account_transactions() failed with {e} current balance={await cls.get_balance(account_id)}"
                    )
                    raise



from typing import Dict

from liualgotrader.common import config
from liualgotrader.common.types import BrokerType
from liualgotrader.trading.alpaca import AlpacaTrader
from liualgotrader.trading.base import Trader
from liualgotrader.trading.gemini import GeminiTrader
from liualgotrader.trading.tradier import TradierTrader

traders: Dict[str, Trader] = {}


def trader_factory(*args, **kwargs) -> Trader:
    global traders

    if config.broker == BrokerType.alpaca:
        if "ALPACA" not in traders:
            traders["ALPACA"] = AlpacaTrader(*args, **kwargs)
        return traders["ALPACA"]
    elif config.broker == BrokerType.gemini:
        if "GEMINI" not in traders:
            traders["GEMINI"] = GeminiTrader(*args, **kwargs)
        return traders["GEMINI"]
    elif config.broker == BrokerType.tradier:
        if "TRADIER" not in traders:
            traders["TRADIER"] = TradierTrader(*args, **kwargs)
        return traders["TradierTrader"]
    else:
        raise Exception(f"unsupported broker  {config.broker}")


def get_trader_by_name(trader_name: str) -> Trader:
    global traders

    if trader_name not in traders:
        raise ValueError(f"Trader {trader_name} was not initialized")

    return traders[trader_name]



import asyncpg
import pandas as pd

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog


async def create_db_connection(dsn: str = None) -> None:
    # if not hasattr(config, "db_conn_pool"):
    _dsn = dsn or config.dsn
    config.db_conn_pool = await asyncpg.create_pool(
        dsn=_dsn,
        min_size=2,
        max_size=10,
    )

    tlog(f"db connection pool initialized w {_dsn}")


async def fetch_as_dataframe(query: str, *args) -> pd.DataFrame:
    try:
        config.db_conn_pool
    except (NameError, AttributeError):
        await create_db_connection()

    print("db_conn_pool", config.db_conn_pool, id(config.db_conn_pool))
    async with config.db_conn_pool.acquire() as con:
        print("connection object", con, id(con))
        stmt = await con.prepare(query)
        columns = [a.name for a in stmt.get_attributes()]
        data = await stmt.fetch(*args)

        return (
            pd.DataFrame(data=data, columns=columns)
            if data and len(data) > 0
            else pd.DataFrame()
        )



import asyncio
import uuid
from datetime import date, datetime, timedelta
from typing import List, Optional, Tuple

import pandas as pd

from liualgotrader.common.types import Order, QueueMapper
from liualgotrader.models.algo_run import AlgoRun


class Trader:
    __instance: object = None

    def __init__(self, queues: QueueMapper = None):
        self.queues = queues
        Trader.__instance = self

    def __repr__(self):
        return type(self).__name__

    async def create_session(self, algo_name: str) -> AlgoRun:
        new_batch_id = str(uuid.uuid4())
        algo_run = AlgoRun(algo_name, new_batch_id)
        await algo_run.save()
        return algo_run

    def get_market_schedule(
        self,
    ) -> Tuple[Optional[datetime], Optional[datetime]]:
        ...

    def get_trading_days(
        self, start_date: date, end_date: date = date.today()
    ) -> pd.DataFrame:
        ...

    def is_market_open_today(self) -> bool:
        ...

    def is_market_open(self, now: datetime) -> bool:
        if not hasattr(self, "market_open"):
            self.market_open, self.market_close = self.get_market_schedule()  # type: ignore
        return (
            False
            if not self.market_open or not self.market_close
            else self.market_open <= now <= self.market_close
        )

    def get_time_market_close(self) -> Optional[timedelta]:
        ...

    def get_position(self, symbol: str) -> float:
        ...

    async def reconnect(self):
        ...

    async def get_tradeable_symbols(self) -> List[str]:
        ...

    async def get_shortable_symbols(self) -> List[str]:
        ...

    async def is_shortable(self, symbol: str) -> bool:
        ...

    async def is_order_completed(
        self, order_id: str, external_order_id: Optional[str] = None
    ) -> Tuple[
        Order.EventType, Optional[float], Optional[float], Optional[float]
    ]:
        """return filled order status, average filled price, filled quantity and trade fee"""
        ...

    async def is_fractionable(self, symbol: str) -> bool:
        ...

    async def submit_order(
        self,
        symbol: str,
        qty: float,
        side: str,
        order_type: str,
        time_in_force: str,
        limit_price: str = None,
        stop_price: str = None,
        client_order_id: str = None,
        extended_hours: bool = None,
        order_class: str = None,
        take_profit: dict = None,
        stop_loss: dict = None,
        trail_price: str = None,
        trail_percent: str = None,
        on_behalf_of: str = None,
    ) -> Order:
        ...

    async def get_order(self, order_id: str) -> Order:
        ...

    async def get_account_order(
        self, external_account_id: str, order_id: str
    ) -> Order:
        ...

    async def cancel_order(self, order: Order) -> bool:
        ...

    async def run(self) -> Optional[asyncio.Task]:
        ...

    async def close(self):
        ...

    @classmethod
    def get_instance(cls):
        if not cls.__instance:
            raise AssertionError("Must instantiate before usage")

        return cls.__instance  # type: ignore



# type: ignore
import concurrent.futures
from datetime import date, datetime, timedelta
from typing import Dict, List, Optional, Tuple

import alpaca_trade_api as tradeapi
import pandas as pd
from dateutil.parser import parse as date_parser
from pytz import timezone

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog, tlog_exception
from liualgotrader.common.types import DataConnectorType, TimeScale
from liualgotrader.data.data_base import DataAPI
from liualgotrader.data.data_factory import data_loader_factory

nyc = timezone("America/New_York")

m_and_a_data = pd.read_csv(
    "https://raw.githubusercontent.com/amor71/LiuAlgoTrader/master/database/market_m_a_data.csv"
).set_index("date")


def _calc_data_to_fetch(s: slice, index: pd.Index) -> List[slice]:
    if index.empty:
        return [s]

    slices = []
    if s.start.date() < index[0].date():
        slices.append(slice(s.start, index[0]))

    if s.stop > index[-1]:
        slices.append(slice(index[-1], s.stop))

    return slices


def convert_offset_to_datetime(
    data_api: DataAPI,
    symbol: str,
    index: pd.Index,
    scale: TimeScale,
    offset: int,
    start: Optional[datetime] = None,
) -> datetime:
    try:
        return index[offset]
    except IndexError:
        last_trading_time = start or data_api.get_last_trading(symbol)
        if scale == TimeScale.minute:
            return last_trading_time.replace(
                second=0, microsecond=0
            ) + timedelta(minutes=1 + offset)

        return data_api.get_trading_day(symbol, last_trading_time, 1 + offset)


def handle_slice_conversion(
    data_api: DataAPI,
    symbol: str,
    key: slice,
    scale: TimeScale,
    index: pd.Index,
) -> slice:
    # handle slice end
    if type(key.stop) == str:
        key = slice(
            key.start,
            nyc.localize(date_parser(key.stop)),
        )
    elif type(key.stop) == int:
        key = slice(
            key.start,
            convert_offset_to_datetime(
                data_api, symbol, index, scale, key.stop
            ),
        )
    elif type(key.stop) == date:
        key = slice(
            key.start,
            nyc.localize(datetime.combine(key.stop, datetime.min.time())),
        )
    elif type(key.stop) == datetime and key.stop.tzinfo is None:
        key = slice(key.start, nyc.localize(key.stop))

    # handle slide start
    if type(key.start) == str:
        key = slice(nyc.localize(date_parser(key.start)), key.stop)
    elif type(key.start) == int:
        key = slice(
            convert_offset_to_datetime(
                data_api, symbol, index, scale, key.start, key.stop
            ),
            key.stop,
        )
    elif type(key.start) == date:
        key = slice(
            nyc.localize(datetime.combine(key.start, datetime.min.time())),
            key.stop,
        )
    elif type(key.start) == datetime and key.start.tzinfo is None:
        key = slice(nyc.localize(key.start), key.stop)

    return key


def load_item_by_offset(
    data_api: DataAPI,
    symbol_data: pd.DataFrame,
    symbol: str,
    scale: TimeScale,
    offset: int,
    concurrency: int,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    i = convert_offset_to_datetime(
        data_api=data_api,
        symbol=symbol,
        index=symbol_data.index,
        scale=scale,
        offset=offset,
    )
    symbol_data = fetch_data_datetime(
        data_api=data_api,
        symbol_data=symbol_data,
        symbol=symbol,
        scale=scale,
        d=i,
        concurrency=concurrency,
    )
    # index = symbol_data.index.get_indexer([i], method="nearest")[0]
    return (
        symbol_data,
        symbol_data.iloc[offset],  # index
    )


def get_item_by_offset(
    data_api: DataAPI,
    symbol_data: pd.DataFrame,
    symbol: str,
    scale: TimeScale,
    offset: int,
    concurrency: int,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    try:
        return symbol_data, symbol_data.iloc[len(symbol_data.index) + offset]
    except IndexError:
        return load_item_by_offset(
            data_api, symbol_data, symbol, scale, offset, concurrency
        )


def _data_fetch_executor(
    data_api: DataAPI,
    symbol: str,
    scale: TimeScale,
    start: datetime,
    end: datetime,
) -> pd.DataFrame:

    df = data_api.get_symbol_data(
        symbol,
        start=(start.date() if isinstance(start, datetime) else start),
        end=(end.date() if isinstance(end, datetime) else end)
        + timedelta(days=1),
        scale=scale,
    )

    return df.reindex(
        columns=[
            "open",
            "high",
            "low",
            "close",
            "volume",
            "vwap",
            "average",
            "count",
        ]
    ).sort_index()


def _concurrent_fetch_data(
    data_api: DataAPI,
    symbol_data: pd.DataFrame,
    symbol: str,
    scale: TimeScale,
    start: datetime,
    end: datetime,
):
    ranges = data_api.data_concurrency_ranges(
        symbol=symbol, start=start, end=end, scale=scale
    )
    if not len(ranges):
        raise ValueError(f"can't load empty range list {ranges}")

    ranges = list(zip(ranges, ranges[1:]))
    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = {
            executor.submit(
                _data_fetch_executor,
                data_api,
                symbol,
                scale,
                range[0],
                range[1],
            ): range
            for range in ranges
        }
        for future in concurrent.futures.as_completed(futures):
            response = future.result()
            symbol_data = pd.concat([symbol_data, response])
            symbol_data = symbol_data[
                ~symbol_data.index.duplicated(keep="first")
            ].sort_index()

    return symbol_data.reindex(
        columns=[
            "open",
            "high",
            "low",
            "close",
            "volume",
            "vwap",
            "average",
            "count",
        ]
    ).sort_index()


def _legacy_fetch_data_range(
    data_api: DataAPI,
    symbol_data: pd.DataFrame,
    symbol: str,
    scale: TimeScale,
    start: datetime,
    end: datetime,
) -> pd.DataFrame:

    adjusted_symbol = symbol
    m_and_a_data.index = m_and_a_data.index.astype("datetime64[ns]", copy=True)
    while True:
        adjusted_data = m_and_a_data.loc[
            (
                end.replace(
                    hour=0, minute=0, second=0, microsecond=0, tzinfo=None
                )
                <= m_and_a_data.index
            )
            & (m_and_a_data.to_symbol == adjusted_symbol)
        ]
        if not adjusted_data.empty:
            adjusted_symbol = adjusted_data.from_symbol.item()
        else:
            break

    if adjusted_symbol != symbol:
        adjusted_df = data_api.get_symbol_data(
            adjusted_symbol,
            start=(start.date() if isinstance(start, datetime) else start),
            end=(end.date() if isinstance(end, datetime) else end)
            + timedelta(days=1),
            scale=scale,
        )
    else:
        adjusted_df = pd.DataFrame()

    new_df = data_api.get_symbol_data(
        symbol,
        start=(start.date() if isinstance(start, datetime) else start),
        end=(end.date() if isinstance(end, datetime) else end)
        + timedelta(days=1),
        scale=scale,
    )
    symbol_data = pd.concat([adjusted_df, new_df, symbol_data], sort=True)
    symbol_data = symbol_data[~symbol_data.index.duplicated(keep="first")]
    return symbol_data.reindex(
        columns=[
            "open",
            "high",
            "low",
            "close",
            "volume",
            "vwap",
            "average",
            "count",
        ]
    ).sort_index()


def fetch_data_range(
    data_api: DataAPI,
    symbol_data: pd.DataFrame,
    symbol: str,
    scale: TimeScale,
    start: datetime,
    end: datetime,
    concurrency: int,
) -> pd.DataFrame:
    if concurrency:
        return _concurrent_fetch_data(
            data_api=data_api,
            symbol_data=symbol_data,
            symbol=symbol,
            scale=scale,
            start=start,
            end=end,
        )
    else:
        return _legacy_fetch_data_range(
            data_api=data_api,
            symbol_data=symbol_data,
            symbol=symbol,
            scale=scale,
            start=start,
            end=end,
        )


def fetch_data_datetime(
    data_api: DataAPI,
    symbol_data: pd.DataFrame,
    symbol: str,
    scale: TimeScale,
    d: datetime,
    concurrency: int,
) -> pd.DataFrame:
    if not symbol_data.empty and d < symbol_data.index.min():
        start = d
        end = symbol_data.index.min()
    elif not symbol_data.empty and d > symbol_data.index.max():
        start = symbol_data.index.max()
        end = d + (
            timedelta(days=1)
            if scale == TimeScale.day
            else timedelta(minutes=1)
        )
    else:
        start = d
        end = d + (
            timedelta(days=1)
            if scale == TimeScale.day
            else timedelta(minutes=1)
        )

    return fetch_data_range(
        data_api=data_api,
        symbol_data=symbol_data,
        symbol=symbol,
        scale=scale,
        start=start,
        end=end,
        concurrency=concurrency,
    )


def getitem_slice(
    data_api: DataAPI,
    symbol: str,
    symbol_data: pd.DataFrame,
    scale: TimeScale,
    key: slice,
    concurrency: int,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    key = slice(key.start or 0, key.stop or -1)

    # ensure key represents datetime
    converted_key = handle_slice_conversion(
        data_api, symbol, key, scale, symbol_data.index
    )

    # load data if needed
    for s in _calc_data_to_fetch(converted_key, symbol_data.index):
        symbol_data = fetch_data_range(
            data_api=data_api,
            symbol_data=symbol_data,
            symbol=symbol,
            scale=scale,
            start=s.start,
            end=s.stop,
            concurrency=concurrency,
        )

    # return data range
    if not isinstance(key.start, int):
        key_start_index = symbol_data.index.get_indexer(
            [converted_key.start], method="nearest"
        )[0]
    elif key.start < 0:
        key_start_index = len(symbol_data.index) + key.start
    else:
        key_start_index = key.start

    if not isinstance(key.stop, int):
        key_end_index = symbol_data.index.get_indexer(
            [converted_key.stop], method="nearest"
        )[0]
    elif key.stop < 0:
        key_end_index = len(symbol_data.index) + key.stop
    else:
        key_end_index = key.stop

    return (
        symbol_data,
        symbol_data.iloc[key_start_index : key_end_index + 1],
    )


def getitem(
    data_api: DataAPI,
    symbol_data: pd.DataFrame,
    symbol: str,
    scale: TimeScale,
    key,
    concurrency,
):
    if type(key) == str:
        key = nyc.localize(date_parser(key))
    elif type(key) == int:
        symbol_data, rc = get_item_by_offset(
            data_api=data_api,
            symbol_data=symbol_data,
            symbol=symbol,
            scale=scale,
            offset=key,
            concurrency=concurrency,
        )
        return symbol_data, rc

    elif type(key) == date:
        key = nyc.localize(datetime.combine(key, datetime.min.time()))
    elif type(key) == datetime and key.tzinfo is None:
        key = nyc.localize(key)

    for s in _calc_data_to_fetch(slice(key, key), symbol_data.index):
        symbol_data = fetch_data_range(
            data_api=data_api,
            symbol_data=symbol_data,
            symbol=symbol,
            scale=scale,
            start=s.start,
            end=s.stop,
            concurrency=concurrency,
        )

    if symbol_data.empty:
        raise ValueError(f"details for symbol {symbol} do not exist")

    index = symbol_data.index.get_indexer([key], method="ffill")[0]
    return symbol_data, symbol_data.iloc[index]


class SymbolData:
    class _Column:
        def __init__(
            self,
            data_api: DataAPI,
            name: str,
            data: object,
            scale: TimeScale,
            concurrency: int,
        ):
            self.name = name
            self.data_api = data_api
            self.data = data
            self.scale = scale
            self.concurrency = concurrency

        def __repr__(self):
            return str(self.data.symbol_data[self.name])

        def _get_index(self, index: datetime, method: str = "ffill") -> int:
            try:
                return self.data.symbol_data.index.get_loc(
                    index, method=method
                )
            except KeyError:
                self.data.symbol_data = fetch_data_datetime(
                    data_api=self.data_api,
                    symbol_data=self.data.symbol_data,
                    symbol=self.data.symbol,
                    scale=self.scale,
                    d=index,
                    concurrency=self.concurrency,
                )
                return self.data.symbol_data.index.get_loc(
                    index, method="nearest"
                )

        def __getitem__(self, key):
            try:
                if type(key) == slice:
                    self.data.symbol_data, rc = getitem_slice(
                        data_api=self.data_api,
                        symbol=self.data.symbol,
                        symbol_data=self.data.symbol_data,
                        scale=self.scale,
                        key=key,
                        concurrency=self.concurrency,
                    )
                    return rc[self.name]

                self.data.symbol_data, rc = getitem(
                    data_api=self.data_api,
                    symbol=self.data.symbol,
                    symbol_data=self.data.symbol_data,
                    scale=self.scale,
                    key=key,
                    concurrency=self.concurrency,
                )
                return rc[self.name]
            except Exception:
                if config.debug_enabled:
                    tlog_exception("__getitem__")
                raise

        def __getattr__(self, attr):
            return self.data.symbol_data[self.name].__getattr__(attr)

        def __call__(self):
            return self.data.symbol_data[self.name]

    def __init__(
        self,
        data_api: tradeapi,
        symbol: str,
        scale: TimeScale,
        concurrency: int,
        prefetched_data: Optional[pd.DataFrame] = None,
    ):
        self.data_api = data_api
        self.symbol = symbol
        self.scale = scale
        self.concurrency = concurrency
        self.columns: Dict[str, self._Column] = {}  # type: ignore

        self.symbol_data = (
            prefetched_data
            if prefetched_data is not None
            else pd.DataFrame(
                columns=[
                    "open",
                    "high",
                    "low",
                    "close",
                    "volume",
                    "vwap",
                    "average",
                    "count",
                ]
            )
        )

    #    def __setattr__(self, name, value):
    #        return self.symbol_data.__setattr__(name, value)

    def __getattr__(self, attr) -> _Column:
        if attr[:3] == "loc" or attr[:4] == "iloc" or attr[:5] == "apply":
            return self.symbol_data.__getattr__(attr)
        elif attr not in self.columns:
            self.columns[attr] = self._Column(
                self.data_api, attr, self, self.scale, self.concurrency
            )
        return self.columns[attr]

    def _get_index(self, index: datetime, method: str = "ffill") -> int:
        try:
            return self.symbol_data.index.get_loc(index, method=method)
        except ValueError:
            tlog(f"[EXCEPTION] ValueError {index},{self.symbol_data.index}")
            raise
        except KeyError:
            self.symbol_data = fetch_data_datetime(
                data_api=self.data_api,
                symbol_data=self.symbol_data,
                symbol=self.symbol,
                scale=self.scale,
                d=index,
                concurrency=self.concurrency,
            )
            return self.data.symbol_data.index.get_loc(index, method="nearest")

    def __getitem__(self, key):
        try:
            if type(key) == slice:
                self.symbol_data, rc = getitem_slice(
                    data_api=self.data_api,
                    symbol=self.symbol,
                    symbol_data=self.symbol_data,
                    scale=self.scale,
                    key=key,
                    concurrency=self.concurrency,
                )
                return rc

            self.symbol_data, rc = getitem(
                data_api=self.data_api,
                symbol=self.symbol,
                symbol_data=self.symbol_data,
                scale=self.scale,
                key=key,
                concurrency=self.concurrency,
            )
            return rc
        except Exception:
            if config.debug_enabled:
                tlog_exception("__getitem__")
            raise

    def __repr__(self):
        return str(self.symbol_data)


class DataLoader:
    def __init__(
        self,
        scale: TimeScale = TimeScale.minute,
        connector: DataConnectorType = config.data_connector,
        concurrency: Optional[int] = 0,
    ):
        self.data_api = data_loader_factory(connector)
        self.data: Dict[str, SymbolData] = {}
        self.scale = scale
        self.concurrency = concurrency
        if not self.data_api:
            raise AssertionError("Failed to create data loader")

    def keys(self) -> List[str]:
        return list(self.data.keys())

    def pre_fetch(self, symbols: List[str], start: date, end: date):
        data = self.data_api.get_symbols_data(
            symbols=symbols, start=start, end=end, scale=self.scale
        )
        for symbol, df in data.items():
            self.data[symbol] = SymbolData(
                self.data_api, symbol, self.scale, self.concurrency, df
            )

    def exist(self, symbol: str) -> bool:
        return symbol in self.data

    def __len__(self) -> int:
        return len(self.data.keys())

    def __getattr__(self, attr):
        return self.__getitem__(attr)

    def __getitem__(self, symbol: str) -> SymbolData:
        if not self.data_api:
            raise AssertionError("Must call a well constructed object")

        if symbol not in self.data:
            self.data[symbol] = SymbolData(
                self.data_api, symbol, self.scale, self.concurrency
            )

        return self.data[symbol]



import asyncio

import pytz

from liualgotrader.analytics.analysis import load_trades_by_portfolio
from liualgotrader.models.accounts import Accounts
from liualgotrader.models.portfolio import Portfolio


async def _calc_account_transactions(portfolio_id: str, account_id: int):
    _df = load_trades_by_portfolio(portfolio_id)
    local = pytz.timezone("UTC")
    for _, row in _df.iterrows():
        utc_dt = local.localize(row.tstamp, is_dst=None)
        await Accounts.add_transaction(
            account_id,
            (row.qty * row.price) * (-1 if row.operation == "buy" else 1),
            utc_dt,
        )


def account_transactions(portfolio_id: str):
    loop = asyncio.get_event_loop()
    _ = loop.run_until_complete(Portfolio.load_by_portfolio_id(portfolio_id))
    account_id, account_size = loop.run_until_complete(
        Portfolio.load_details(portfolio_id)
    )
    loop.run_until_complete(Accounts.clear_balance(account_id, account_size))
    loop.run_until_complete(Accounts.clear_account_transactions(account_id))

    loop.run_until_complete(
        _calc_account_transactions(
            portfolio_id=portfolio_id, account_id=account_id
        )
    )



import os
import sys
import traceback
from datetime import datetime

try:
    from liualgotrader.common.config import gcp_logger

    if gcp_logger:
        from google.cloud import logging

        logger = logging.Client().logger("trader")
    else:
        logger = None
except Exception:
    logger = None


def tlog(msg: str, origin: str = None) -> None:
    try:
        calling_fn = origin or f"[{sys._getframe(1).f_code.co_name}()]"
    except Exception:
        calling_fn = ""

    if logger:
        try:
            logger.log_text(f"{calling_fn}[{os.getpid()}] {msg}")
        except Exception as e:
            print(f"[Error] exception when trying to log to Stackdriver {e}")
    print(f"{calling_fn}[{os.getpid()}]{datetime.now()}:{msg}", flush=True)


def tlog_exception(origin: str):
    traceback.print_exc()
    exc_info = sys.exc_info()
    lines = traceback.format_exception(*exc_info)
    for line in lines:
        tlog(f"{line}", origin)
    del exc_info



import asyncio
import concurrent.futures
import queue
import time
import traceback
from datetime import date, datetime, timedelta
from random import randint
from typing import Callable, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import pandas_market_calendars
import pytz
import requests
from alpaca_trade_api.rest import REST, URL, APIError, TimeFrame
from alpaca_trade_api.stream import Stream
from dateutil.parser import parse as date_parser

from liualgotrader.common import config
from liualgotrader.common.list_utils import chunks
from liualgotrader.common.tlog import tlog, tlog_exception
from liualgotrader.common.types import QueueMapper, TimeScale, WSEventType
from liualgotrader.data.data_base import DataAPI
from liualgotrader.data.streaming_base import StreamingAPI

NY = "America/New_York"
nytz = pytz.timezone(NY)


def _is_crypto_symbol(symbol: str) -> bool:
    return symbol.lower() in {"eth/usd", "btc/usd", "ethusd", "btcusd"}


class AlpacaData(DataAPI):
    def __init__(self):
        self.alpaca_rest_client = REST(
            key_id=config.alpaca_api_key, secret_key=config.alpaca_api_secret
        )
        if not self.alpaca_rest_client:
            raise AssertionError(
                "Failed to authenticate Alpaca RESTful client"
            )
        # for requesting market snapshots by chunk of symbols
        self.symbol_chunk_size = 1000
        self.datetime_cache: Dict[datetime, datetime] = {}

    def get_symbols(self) -> List[str]:
        if not self.alpaca_rest_client:
            raise AssertionError("Must call w/ authenticated Alpaca client")

        return [
            asset.symbol
            for asset in self.alpaca_rest_client.list_assets(
                status="active", asset_class="us_equity"
            )
            if asset.tradable
        ]

    async def get_market_snapshot(
        self, filter_func: Optional[Callable] = None
    ) -> List[Dict]:
        # parse market snapshots per chunk of symbols
        symbols = self.get_symbols()
        return await self._get_symbols_snapshot(symbols, filter_func)

    async def _get_symbols_snapshot(
        self, symbols: List[str], filter_func: Optional[Callable]
    ) -> List[Dict]:
        def _parse_ticker_snapshot(_ticker: str, _ticket_snapshot: object):
            try:
                return {
                    "ticker": _ticker,
                    **{
                        sub_snapshot_type: _sub_snapshot_obj.__dict__["_raw"]
                        for sub_snapshot_type, _sub_snapshot_obj in _ticket_snapshot.__dict__.items()
                    },
                }
            # skip over if some snapshot type is missing (e.g. "prev_daily_bar": None)
            except AttributeError:
                return None

        def _parse_snapshot_and_filter(_symbols: List[str]) -> List[Dict]:
            processed_tickers_snapshot = map(
                lambda key_and_val: _parse_ticker_snapshot(*key_and_val),
                self.alpaca_rest_client.get_snapshots(_symbols).items(),
            )
            return list(
                filter(
                    lambda snapshot: (  # type: ignore
                        (snapshot is not None) and (filter_func(snapshot))
                    ),
                    list(processed_tickers_snapshot),  # type : ignore
                )
                if filter_func is not None
                else processed_tickers_snapshot
            )

        # request snapshots per chunk of tickers by concurrency
        with concurrent.futures.ThreadPoolExecutor() as executor:
            loop = asyncio.get_event_loop()
            futures = [
                loop.run_in_executor(
                    executor,
                    _parse_snapshot_and_filter,
                    symbols[symbol_idx : symbol_idx + self.symbol_chunk_size],
                )
                for symbol_idx in range(
                    0, len(symbols), self.symbol_chunk_size
                )
            ]

            market_snapshots = [
                y for x in await asyncio.gather(*futures) for y in x
            ]

        return market_snapshots

    def _localize_start_end(self, start: date, end: date) -> Tuple[str, str]:
        return (
            nytz.localize(
                datetime.combine(start, datetime.min.time())
            ).isoformat(),
            (
                nytz.localize(
                    datetime.now().replace(microsecond=0)
                ).isoformat()
                if end >= date.today()
                else nytz.localize(
                    datetime.combine(end, datetime.min.time())
                ).isoformat()
            ),
        )

    def get_last_trading(self, symbol: str) -> datetime:
        if not self.alpaca_rest_client:
            raise AssertionError("Must call w/ authenticated Alpaca client")

        if _is_crypto_symbol(symbol):
            return datetime.now(tz=nytz)
        try:
            snapshot_data = self.alpaca_rest_client.get_snapshot(symbol)
        except APIError as e:
            raise ValueError(f"{symbol} snapshot not found") from e

        min_bar = snapshot_data.latest_trade
        if not min_bar:
            raise ValueError(f"Can't get snapshot for {symbol}")

        return min_bar.t

    def get_trading_holidays(self) -> List[str]:
        nyse = pandas_market_calendars.get_calendar("NYSE")
        return nyse.holidays().holidays

    def get_trading_day(
        self, symbol: str, now: datetime, offset: int
    ) -> datetime:
        if _is_crypto_symbol(symbol):
            cbd_offset = timedelta(days=offset)
        else:
            cbd_offset = pd.tseries.offsets.CustomBusinessDay(
                n=offset - 1, holidays=self.get_trading_holidays()
            )
        return (
            nytz.localize(now + cbd_offset)
            if now.tzinfo is None
            else now + cbd_offset
        )

    def num_trading_minutes(self, symbol: str, start: date, end: date) -> int:
        return (24 if _is_crypto_symbol(symbol) else (20 - 4)) * 60

    def num_trading_days(self, symbol: str, start: date, end: date) -> int:
        if type(start) == str:
            start = date_parser(start)  # type: ignore
        if type(end) == str:
            end = date_parser(end)  # type: ignore

        return (
            (end - start).days + 1
            if _is_crypto_symbol(symbol)
            else len(
                pd.date_range(
                    start,
                    end,
                    freq=pd.tseries.offsets.CustomBusinessDay(
                        holidays=self.get_trading_holidays()
                    ),
                )
            )
        )

    def get_max_data_points_per_load(self) -> int:
        # Alpaca suggests 10000 points
        return 10000

    def trading_days_slice(self, symbol: str, s: slice) -> slice:
        if not self.alpaca_rest_client:
            raise AssertionError("Must call w/ authenticated Alpaca client")

        if _is_crypto_symbol(symbol):
            return s

        if s.start in self.datetime_cache and s.stop in self.datetime_cache:
            return slice(
                self.datetime_cache[s.start], self.datetime_cache[s.stop]
            )

        trading_days = self.alpaca_rest_client.get_calendar(
            str(s.start.date()), str(s.stop.date())
        )
        new_slice = slice(
            nytz.localize(
                datetime.combine(
                    trading_days[0].date.date(), trading_days[0].open
                )
            ),
            nytz.localize(
                datetime.combine(
                    trading_days[-1].date.date(), trading_days[-1].open
                )
            ),
        )

        self.datetime_cache[s.start] = new_slice.start
        self.datetime_cache[s.stop] = new_slice.stop

        return new_slice

    def crypto_get_symbol_data(
        self,
        symbol: str,
        start: str,
        end: str,
        timeframe: TimeFrame,
    ) -> pd.DataFrame:
        if "/" not in symbol:
            symbol = f"{symbol[:3]}/{symbol[3:]}"
        symbol = symbol.upper()
        url = f"{config.alpaca_crypto_base_url}/bars"
        page_token = None

        rc_df = pd.DataFrame()

        while True:
            response = requests.get(
                url,
                params={  # type:ignore
                    "symbols": [symbol],
                    "start": start,
                    "end": end,
                    "limit": self.get_max_data_points_per_load(),
                    "timeframe": "1Day"
                    if timeframe == TimeFrame.Day
                    else "1Min",
                    "page_token": page_token,
                },
                headers={
                    "APCA-API-KEY-ID": config.alpaca_api_key,
                    "APCA-API-SECRET-KEY": config.alpaca_api_secret,
                },
            )

            response.raise_for_status()

            json_data = response.json()
            df = pd.DataFrame(json_data["bars"][symbol])
            df.rename(
                columns={
                    "o": "open",
                    "c": "close",
                    "h": "high",
                    "l": "low",
                    "v": "volume",
                    "vw": "vwap",
                    "t": "timestamp",
                    "n": "trade_count",
                },
                inplace=True,
            )
            df["timestamp"] = pd.to_datetime(df.timestamp)
            df = df.set_index(df.timestamp)

            rc_df = pd.concat([rc_df, df], sort=True)
            rc_df = rc_df[~rc_df.index.duplicated(keep="first")]

            page_token = json_data["next_page_token"]
            if page_token is None:
                return rc_df

    def get_symbols_data(
        self,
        symbols: List[str],
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> Dict[str, pd.DataFrame]:
        if not self.alpaca_rest_client:
            raise AssertionError("Must call w/ authenticated Alpaca client")
        if not isinstance(symbols, list):
            raise AssertionError(f"{symbols} must be a list")

        if scale == TimeScale.minute:
            end += timedelta(days=1)
        _start, _end = self._localize_start_end(start, end)
        dfs: Dict = {}
        t: TimeFrame = (
            TimeFrame.Minute
            if scale == TimeScale.minute
            else TimeFrame.Day
            if scale == TimeScale.day
            else None
        )
        try:
            data = self.alpaca_rest_client.get_bars(
                symbol=symbols,
                timeframe=t,
                start=_start,
                end=_end,
                limit=1000000000,
                adjustment="all",
            ).df
        except requests.exceptions.HTTPError as e:
            tlog(f"received HTTPError: {e}")
            if e.response.status_code in (500, 502, 504, 429):
                tlog("retrying")
                time.sleep(10)
                return self.get_symbols_data(symbols, start, end, scale)

        data = data.tz_convert("America/New_York")
        data["average"] = data.vwap
        data["count"] = data.trade_count
        data["vwap"] = np.NaN
        grouped = data.groupby(data.symbol)
        for symbol in data.symbol.unique():
            dfs[symbol] = grouped.get_group(symbol)[
                [
                    "open",
                    "high",
                    "low",
                    "close",
                    "volume",
                    "count",
                    "average",
                    "vwap",
                ]
            ]

        return dfs

    def get_symbol_data(
        self,
        symbol: str,
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> pd.DataFrame:
        _start, _end = self._localize_start_end(start, end)

        if not self.alpaca_rest_client:
            raise AssertionError("Must call w/ authenticated Alpaca client")

        t: TimeFrame = (
            TimeFrame.Minute
            if scale == TimeScale.minute
            else TimeFrame.Day
            if scale == TimeScale.day
            else None
        )

        try:
            if config.detailed_dl_debug_enabled:
                tlog(f"symbol={symbol}, timeframe={t}, range=({_start, _end})")

            data = (
                self.crypto_get_symbol_data(
                    symbol=symbol, start=_start, end=_end, timeframe=t
                )
                if _is_crypto_symbol(symbol)
                else self.alpaca_rest_client.get_bars(
                    symbol=symbol,
                    timeframe=t,
                    start=_start,
                    end=_end,
                    limit=1000000,
                    adjustment="all",
                ).df
            )
        except requests.exceptions.HTTPError as e:
            tlog(f"received HTTPError: {e}")
            if e.response.status_code in (500, 502, 504, 429):
                tlog("retrying")
                time.sleep(10)
                return self.get_symbol_data(symbol, start, end, scale)

        except Exception as e:
            raise ValueError(
                f"[EXCEPTION] {e} for {symbol} has no data for {_start} to {_end} w {scale.name}"
            )
        else:
            if data.empty:
                raise ValueError(
                    f"[ERROR] {symbol} has no data for {_start} to {_end} w {scale.name}"
                )

        data.index = data.index.tz_convert("America/New_York")
        data["average"] = data.vwap
        data["count"] = data.trade_count
        data["vwap"] = np.NaN

        return data[
            [
                "open",
                "high",
                "low",
                "close",
                "volume",
                "count",
                "average",
                "vwap",
            ]
        ]


class AlpacaStream(StreamingAPI):
    def __init__(self, queues: QueueMapper):
        self.alpaca_ws_client = Stream(
            base_url=URL(config.alpaca_base_url),
            key_id=config.alpaca_api_key,
            secret_key=config.alpaca_api_secret,
            data_feed=config.alpaca_data_feed,
        )

        if not self.alpaca_ws_client:
            raise AssertionError(
                "Failed to authenticate Alpaca web_socket client"
            )

        self.task: Optional[asyncio.Task] = None
        super().__init__(queues)

    async def run(self):
        if not self.task:
            if self.queues:
                self.task = asyncio.create_task(
                    self.alpaca_ws_client._run_forever()
                )
            else:
                raise AssertionError(
                    "can't call `AlpacaStream.run()` without queues"
                )

    @classmethod
    async def bar_handler(cls, msg):
        try:
            event = {
                "symbol": msg.symbol,
                "open": msg.open,
                "close": msg.close,
                "high": msg.high,
                "low": msg.low,
                "timestamp": pd.to_datetime(
                    msg.timestamp, utc=True
                ).astimezone(nytz),
                "volume": msg.volume,
                "count": int(msg.trade_count),
                "vwap": np.nan,
                "average": msg.vwap,
                "totalvolume": None,
                "EV": "AM",
            }
            cls.get_instance().queues[msg.symbol].put(event, timeout=1)
        except queue.Full as f:
            tlog(
                f"[EXCEPTION] process_message(): queue for {event['sym']} is FULL:{f}"
            )
            raise
        except Exception as e:
            tlog(
                f"[EXCEPTION] process_message(): exception of type {type(e).__name__} with args {e.args}"
            )
            if config.debug_enabled:
                traceback.print_exc()

    @classmethod
    async def crypto_bar_handler(cls, msg):
        try:
            if msg.exchange != "CBSE":
                return

            event = {
                "symbol": msg.symbol,
                "open": msg.open,
                "close": msg.close,
                "high": msg.high,
                "low": msg.low,
                "timestamp": pd.to_datetime(
                    msg.timestamp, utc=True
                ).astimezone(nytz),
                "volume": msg.volume,
                "count": int(msg.trade_count),
                "vwap": np.nan,
                "average": msg.vwap,
                "totalvolume": None,
                "EV": "AM",
            }
            cls.get_instance().queues[msg.symbol].put(event, timeout=1)
        except queue.Full as f:
            tlog(
                f"[EXCEPTION] process_message(): queue for {event['sym']} is FULL:{f}"
            )
            raise
        except Exception as e:
            tlog(
                f"[EXCEPTION] process_message(): exception of type {type(e).__name__} with args {e.args}"
            )
            if config.debug_enabled:
                traceback.print_exc()

    @classmethod
    async def trades_handler(cls, msg):
        try:
            ts = pd.to_datetime(msg.timestamp)
            if (time_diff := (datetime.now(tz=nytz) - ts)) > timedelta(
                seconds=10
            ) and randint(  # nosec
                1, 100
            ) == 1:  # nosec
                tlog(
                    f"Received trade for {msg.symbol} too out of sync w {time_diff}"
                )

            event = {
                "symbol": msg.symbol,
                "price": msg.price,
                "open": msg.price,
                "close": msg.price,
                "high": msg.price,
                "low": msg.price,
                "timestamp": ts,
                "volume": msg.size,
                "exchange": msg.exchange,
                "conditions": msg.conditions
                if hasattr(msg, "conditions")
                else None,
                "tape": msg.tape if hasattr(msg, "tape") else None,
                "average": np.nan,
                "count": 1,
                "vwap": np.nan,
                "EV": "T",
            }

            cls.get_instance().queues[msg.symbol].put(event, block=False)

        except queue.Full as f:
            tlog(
                f"[EXCEPTION] process_message(): queue for {event['sym']} is FULL:{f}"
            )
            raise
        except Exception as e:
            tlog(
                f"[EXCEPTION] process_message(): exception of type {type(e).__name__} with args {e.args}"
            )
            if config.debug_enabled:
                traceback.print_exc()

    @classmethod
    async def crypto_trades_handler(cls, msg):
        try:
            if msg.exchange != "CBSE":
                return

            ts = pd.to_datetime(msg.timestamp)
            if (time_diff := (datetime.now(tz=nytz) - ts)) > timedelta(
                seconds=10
            ) and randint(  # nosec
                1, 100
            ) == 1:  # nosec
                tlog(
                    f"Received trade for {msg.symbol} too out of sync w {time_diff}"
                )

            event = {
                "symbol": msg.symbol,
                "price": msg.price,
                "open": msg.price,
                "close": msg.price,
                "high": msg.price,
                "low": msg.price,
                "timestamp": ts,
                "volume": msg.size,
                "exchange": msg.exchange,
                "conditions": msg.conditions
                if hasattr(msg, "conditions")
                else None,
                "tape": msg.tape if hasattr(msg, "tape") else None,
                "average": np.nan,
                "count": 1,
                "vwap": np.nan,
                "EV": "T",
            }

            cls.get_instance().queues[msg.symbol].put(event, block=False)

        except queue.Full as f:
            tlog(
                f"[EXCEPTION] process_message(): queue for {event['sym']} is FULL:{f}"
            )
            raise
        except Exception as e:
            tlog(
                f"[EXCEPTION] process_message(): exception of type {type(e).__name__} with args {e.args}"
            )
            if config.debug_enabled:
                traceback.print_exc()

    @classmethod
    async def quotes_handler(cls, msg):
        pass

    async def subscribe(
        self, symbols: List[str], events: List[WSEventType]
    ) -> bool:
        tlog(f"Starting subscription for {len(symbols)} symbols")
        upper_symbols = [symbol.upper() for symbol in symbols]
        for syms in chunks(upper_symbols, 1000):
            tlog(f"\tsubscribe {len(syms)}/{len(upper_symbols)}")

            crypto_symbols = list(filter(_is_crypto_symbol, syms))
            equity_symbols = [x for x in syms if x not in crypto_symbols]

            for event in events:
                if event == WSEventType.MIN_AGG:
                    self.alpaca_ws_client._data_ws._running = False

                    if crypto_symbols:
                        self.alpaca_ws_client.subscribe_crypto_bars(
                            AlpacaStream.crypto_bar_handler,
                            *crypto_symbols,
                        )
                    if equity_symbols:
                        self.alpaca_ws_client.subscribe_bars(
                            AlpacaStream.bar_handler,
                            *equity_symbols,
                        )
                elif event == WSEventType.TRADE:
                    if crypto_symbols:
                        self.alpaca_ws_client.subscribe_crypto_trades(
                            AlpacaStream.crypto_trades_handler, *crypto_symbols
                        )
                    if equity_symbols:
                        self.alpaca_ws_client.subscribe_trades(
                            AlpacaStream.trades_handler, *equity_symbols
                        )
                elif event == WSEventType.QUOTE:
                    if crypto_symbols:
                        self.alpaca_ws_client.subscribe_crypto_quotes(
                            AlpacaStream.quotes_handler, *crypto_symbols
                        )
                    if equity_symbols:
                        self.alpaca_ws_client.subscribe_quotes(
                            AlpacaStream.quotes_handler, *equity_symbols
                        )

            await asyncio.sleep(1)

        tlog(f"Completed subscription for {len(symbols)} symbols")
        return True

    async def close(self) -> None:
        tlog("Closing AlpacaStream")

        if self.task:
            # self.alpaca_ws_client.stop()

            await self.alpaca_ws_client.stop_ws()

            while not self.task.done():
                await asyncio.sleep(1.0)

            tlog("Task Done. Closed AlpacaStream")



import asyncio
import os
import queue
import time
import traceback
from datetime import date, datetime, timedelta
from typing import Dict, List, Optional, Tuple

import pandas as pd
import requests
from alpaca_trade_api.entity import Order as AlpacaOrder
from alpaca_trade_api.rest import REST, URL, Entity
from alpaca_trade_api.stream import Stream
from pytz import timezone
from requests.auth import HTTPBasicAuth

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog
from liualgotrader.common.types import Order, QueueMapper, Trade
from liualgotrader.trading.base import Trader

nyc = timezone("America/New_York")


class AlpacaTrader(Trader):
    def __init__(self, qm: QueueMapper = None):
        self.market_open: Optional[datetime]
        self.market_close: Optional[datetime]
        self.alpaca_brokage_api_baseurl = os.getenv(
            "ALPACA_BROKER_API_BASEURL", None
        )
        self.alpaca_brokage_api_key = os.getenv("ALPACA_BROKER_API_KEY", None)
        self.alpaca_brokage_api_secret = os.getenv(
            "ALPACA_BROKER_API_SECRET", None
        )

        self.alpaca_rest_client = REST(
            base_url=URL(config.alpaca_base_url),
            key_id=config.alpaca_api_key,
            secret_key=config.alpaca_api_secret,
        )
        if qm:
            self.alpaca_ws_client = Stream(
                base_url=URL(config.alpaca_base_url),
                key_id=config.alpaca_api_key,
                secret_key=config.alpaca_api_secret,
            )
            if not self.alpaca_ws_client:
                raise AssertionError(
                    "Failed to authenticate Alpaca web_socket client"
                )
            self.alpaca_ws_client.subscribe_trade_updates(
                AlpacaTrader.trade_update_handler
            )
        self.running_task: Optional[asyncio.Task] = None

        now = datetime.now(nyc)
        calendar = self.alpaca_rest_client.get_calendar(
            start=now.strftime("%Y-%m-%d"), end=now.strftime("%Y-%m-%d")
        )[0]

        if now.date() >= calendar.date.date():
            self.market_open = now.replace(
                hour=calendar.open.hour,
                minute=calendar.open.minute,
                second=0,
                microsecond=0,
            )
            self.market_close = now.replace(
                hour=calendar.close.hour,
                minute=calendar.close.minute,
                second=0,
                microsecond=0,
            )
        else:
            self.market_open = self.market_close = None
        super().__init__(qm)

    async def _is_personal_order_completed(
        self, order_id: str
    ) -> Tuple[Order.EventType, float, float, float]:
        alpaca_order = self.alpaca_rest_client.get_order(order_id=order_id)
        event = (
            Order.EventType.canceled
            if alpaca_order.status in ["canceled", "expired", "replaced"]
            else Order.EventType.pending
            if alpaca_order.status in ["pending_cancel", "pending_replace"]
            else Order.EventType.fill
            if alpaca_order.status == "filled"
            else Order.EventType.partial_fill
            if alpaca_order.status == "partially_filled"
            else Order.EventType.other
        )
        return (
            event,
            float(alpaca_order.filled_avg_price or 0.0),
            float(alpaca_order.filled_qty or 0.0),
            0.0,
        )

    async def is_fractionable(self, symbol: str) -> bool:
        try:
            asset_details = self.alpaca_rest_client.get_asset(symbol)
        except Exception:
            return False

        return asset_details.fractionable

    async def _is_brokerage_account_order_completed(
        self, order_id: str, external_order_id: Optional[str] = None
    ) -> Tuple[Order.EventType, float, float, float]:
        if not self.alpaca_brokage_api_baseurl:
            raise AssertionError(
                "order_on_behalf can't be called, if brokerage configs incomplete"
            )

        endpoint: str = (
            f"/v1/trading/accounts/{external_order_id}/orders/{order_id}"
        )
        tlog(f"_is_brokerage_account_order_completed:{endpoint}")
        url: str = self.alpaca_brokage_api_baseurl + endpoint

        response = await self._get_request(url)
        tlog(f"_is_brokerage_account_order_completed: response: {response}")
        event = (
            Order.EventType.canceled
            if response["status"] in ["canceled", "expired", "replaced"]
            else Order.EventType.pending
            if response["status"] in ["pending_cancel", "pending_replace"]
            else Order.EventType.fill
            if response["status"] == "filled"
            else Order.EventType.partial_fill
            if response["status"] == "partially_filled"
            else Order.EventType.other
        )
        return (
            event,
            float(response.get("filled_avg_price") or 0.0),
            float(response.get("filled_qty") or 0.0),
            0.0,
        )

    async def is_order_completed(
        self, order_id: str, external_order_id: Optional[str] = None
    ) -> Tuple[Order.EventType, float, float, float]:
        return (
            await self._is_brokerage_account_order_completed(
                order_id, external_order_id
            )
            if external_order_id
            else await self._is_personal_order_completed(order_id)
        )

    def get_market_schedule(
        self,
    ) -> Tuple[Optional[datetime], Optional[datetime]]:
        return self.market_open, self.market_close

    def get_trading_days(
        self, start_date: date, end_date: date = date.today()
    ) -> pd.DataFrame:
        calendars = self.alpaca_rest_client.get_calendar(
            start=str(start_date), end=str(end_date)
        )
        _df = pd.DataFrame.from_dict([calendar._raw for calendar in calendars])
        _df["date"] = pd.to_datetime(_df.date)
        return _df.set_index("date")

    def get_position(self, symbol: str) -> float:
        pos = self.alpaca_rest_client.get_position(symbol)

        return float(pos.qty) if pos.side == "long" else -1.0 * float(pos.qty)

    def to_order(self, alpaca_order: AlpacaOrder) -> Order:
        event = (
            Order.EventType.canceled
            if alpaca_order.status in ["canceled", "expired", "replaced"]
            else Order.EventType.pending
            if alpaca_order.status in ["pending_cancel", "pending_replace"]
            else Order.EventType.fill
            if alpaca_order.status == "filled"
            else Order.EventType.partial_fill
            if alpaca_order.status == "partially_filled"
            else Order.EventType.other
        )
        return Order(
            order_id=alpaca_order.id,
            symbol=alpaca_order.symbol.lower(),
            event=event,
            price=float(alpaca_order.limit_price or 0.0),
            side=Order.FillSide[alpaca_order.side],
            filled_qty=float(alpaca_order.filled_qty),
            remaining_amount=float(alpaca_order.qty)
            - float(alpaca_order.filled_qty),
            submitted_at=alpaca_order.submitted_at,
            avg_execution_price=alpaca_order.filled_avg_price,
            trade_fees=0.0,
        )

    def _json_to_order(
        self,
        brokerage_response: dict,
        external_account_id: Optional[str] = None,
    ) -> Order:
        event = (
            Order.EventType.canceled
            if brokerage_response["status"]
            in ["canceled", "expired", "replaced"]
            else Order.EventType.pending
            if brokerage_response["status"]
            in ["pending_cancel", "pending_replace"]
            else Order.EventType.fill
            if brokerage_response["status"] == "filled"
            else Order.EventType.partial_fill
            if brokerage_response["status"] == "partially_filled"
            else Order.EventType.other
        )
        return Order(
            order_id=brokerage_response["id"],
            symbol=brokerage_response["symbol"].lower(),
            event=event,
            price=float(brokerage_response["limit_price"] or 0.0),
            side=Order.FillSide[brokerage_response["side"]],
            filled_qty=float(brokerage_response["filled_qty"]),
            remaining_amount=float(brokerage_response["qty"])
            - float(brokerage_response["filled_qty"]),
            submitted_at=pd.Timestamp(
                ts_input=brokerage_response["submitted_at"],
                unit="ms",
                tz="US/Eastern",
            ),
            avg_execution_price=brokerage_response["filled_avg_price"],
            trade_fees=0.0,
            external_account_id=external_account_id,
        )

    async def get_order(self, order_id: str) -> Order:
        return self.to_order(self.alpaca_rest_client.get_order(order_id))

    def is_market_open_today(self) -> bool:
        return self.market_open is not None

    def get_time_market_close(self) -> Optional[timedelta]:
        if not self.is_market_open_today():
            raise AssertionError("Market closed today")

        return (
            self.market_close - datetime.now(nyc)
            if self.market_close
            else None
        )

    async def reconnect(self):
        self.alpaca_rest_client = REST(
            key_id=config.alpaca_api_key, secret_key=config.alpaca_api_secret
        )

    async def run(self) -> asyncio.Task:
        if not self.running_task:
            tlog("starting Alpaca listener")
            self.running_task = asyncio.create_task(
                self.alpaca_ws_client._trading_ws._run_forever()
            )
        return self.running_task

    async def close(self):
        if not self.alpaca_ws_client:
            raise AssertionError("Must call w/ authenticated Alpaca client")
        if self.running_task:
            await self.alpaca_ws_client.stop_ws()

    async def get_tradeable_symbols(self) -> List[str]:
        data = self.alpaca_rest_client.list_assets()
        return [asset.symbol.lower() for asset in data if asset.tradable]

    async def get_shortable_symbols(self) -> List[str]:
        data = self.alpaca_rest_client.list_assets()
        return [
            asset.symbol.lower()
            for asset in data
            if asset.tradable and asset.easy_to_borrow and asset.shortable
        ]

    async def is_shortable(self, symbol) -> bool:
        asset = self.alpaca_rest_client.get_asset(symbol.upper())
        return (
            asset.tradable is not False
            and asset.shortable is not False
            and asset.status != "inactive"
            and asset.easy_to_borrow is not False
        )

    async def _cancel_personal_order(self, order_id: str) -> bool:
        self.alpaca_rest_client.cancel_order(order_id)
        return True

    async def _cancel_brokerage_order(
        self, account_id: str, order_id: str
    ) -> bool:
        if not self.alpaca_brokage_api_baseurl:
            raise AssertionError(
                "_cancel_brokerage_order can't be called, if brokerage configs incomplete"
            )

        endpoint: str = f"/v1/trading/accounts/{account_id}/orders/{order_id}"
        url: str = self.alpaca_brokage_api_baseurl + endpoint

        response_code = await self._delete_request(url)
        tlog(
            f"cancel_brokerage_order {account_id},{order_id} -> {response_code}"
        )
        return response_code == 204

    async def cancel_order(self, order: Order) -> bool:
        if order.external_account_id:
            return await self._cancel_brokerage_order(
                order.external_account_id, order.order_id
            )

        return await self._cancel_personal_order(order.order_id)

    async def _personal_submit(
        self,
        symbol: str,
        qty: float,
        side: str,
        order_type: str,
        time_in_force: str,
        limit_price: str = None,
        stop_price: str = None,
        client_order_id: str = None,
        extended_hours: bool = None,
        order_class: str = None,
        take_profit: dict = None,
        stop_loss: dict = None,
        trail_price: str = None,
        trail_percent: str = None,
        on_behalf_of: str = None,
    ) -> Order:
        o = self.alpaca_rest_client.submit_order(
            symbol.upper(),
            str(qty),
            side,
            order_type,
            time_in_force,
            limit_price,
            stop_price,
            client_order_id,
            extended_hours,
            order_class,
            take_profit,
            stop_loss,
            trail_price,
            trail_percent,
        )

        return self.to_order(o)

    async def _post_request(self, url: str, payload: Dict) -> Dict:
        response = requests.post(
            url=url,
            json=payload,
            auth=HTTPBasicAuth(
                self.alpaca_brokage_api_key, self.alpaca_brokage_api_secret
            ),
        )

        if response.status_code in (429, 504):
            if "x-ratelimit-reset" in response.headers:
                tlog(
                    f"ALPACA BROKERAGE rate-limit till {response.headers['x-ratelimit-reset']}"
                )
                await asyncio.sleep(
                    int(time.time())
                    - int(response.headers["x-ratelimit-reset"])
                )
                tlog("ALPACA BROKERAGE going to retry")
            else:
                tlog(
                    f"ALPACA BROKERAGE push-back w/ {response.status_code} and no x-ratelimit-reset header"
                )
                await asyncio.sleep(10.0)

            return await self._post_request(url, payload)

        if response.status_code in (200, 201, 204):
            return response.json()

        raise AssertionError(
            f"HTTP ERROR {response.status_code} from ALPACA BROKERAGE API with error {response.text}"
        )

    async def _get_request(self, url: str) -> Dict:
        response = requests.get(
            url=url,
            auth=HTTPBasicAuth(
                self.alpaca_brokage_api_key, self.alpaca_brokage_api_secret
            ),
        )

        if response.status_code in (429, 504):
            if "x-ratelimit-reset" in response.headers:
                tlog(
                    f"ALPACA BROKERAGE rate-limit till {response.headers['x-ratelimit-reset']}"
                )
                await asyncio.sleep(
                    int(time.time())
                    - int(response.headers["x-ratelimit-reset"])
                )
                tlog("ALPACA BROKERAGE going to retry")
            else:
                tlog(
                    f"ALPACA BROKERAGE push-back w/ {response.status_code} and no x-ratelimit-reset header"
                )
                await asyncio.sleep(10.0)

            return await self._get_request(url)

        if response.status_code in (200, 201, 204):
            return response.json()

        raise AssertionError(
            f"HTTP ERROR {response.status_code} from ALPACA BROKERAGE API with error {response.text}"
        )

    async def _delete_request(self, url: str) -> int:
        response = requests.delete(
            url=url,
            auth=HTTPBasicAuth(
                self.alpaca_brokage_api_key, self.alpaca_brokage_api_secret
            ),
        )
        # TODO: create a decorator the the re-try / push-backs from server instead of copying.
        if response.status_code in (429, 504):
            if "x-ratelimit-reset" in response.headers:
                tlog(
                    f"ALPACA BROKERAGE rate-limit till {response.headers['x-ratelimit-reset']}"
                )
                await asyncio.sleep(
                    int(time.time())
                    - int(response.headers["x-ratelimit-reset"])
                )
                tlog("ALPACA BROKERAGE going to retry")
            else:
                tlog(
                    f"ALPACA BROKERAGE push-back w/ {response.status_code} and no x-ratelimit-reset header"
                )
                await asyncio.sleep(10.0)

            return await self._delete_request(url)

        return response.status_code

    async def _order_on_behalf(
        self,
        symbol: str,
        qty: float,
        side: str,
        order_type: str,
        time_in_force: str,
        limit_price: str = None,
        stop_price: str = None,
        client_order_id: str = None,
        extended_hours: bool = None,
        order_class: str = None,
        take_profit: dict = None,
        stop_loss: dict = None,
        trail_price: str = None,
        trail_percent: str = None,
        on_behalf_of: str = None,
    ) -> Order:
        if not self.alpaca_brokage_api_baseurl:
            raise AssertionError(
                "order_on_behalf can't be called, if brokerage configs incomplete"
            )

        endpoint: str = f"/v1/trading/accounts/{on_behalf_of}/orders"
        url: str = self.alpaca_brokage_api_baseurl + endpoint

        payload = {
            "symbol": symbol.upper(),
            "qty": qty,
            "side": side,
            "type": order_type,
        }

        if limit_price:
            payload["limit_price"] = limit_price
        if time_in_force:
            payload["time_in_force"] = time_in_force

        json_response: Dict = await self._post_request(
            url=url, payload=payload
        )
        tlog(f"ALPACA BROKERAGE RESPONSE: {json_response}")

        return self._json_to_order(json_response, on_behalf_of)

    async def submit_order(
        self,
        symbol: str,
        qty: float,
        side: str,
        order_type: str,
        time_in_force: str = "day",
        limit_price: str = None,
        stop_price: str = None,
        client_order_id: str = None,
        extended_hours: bool = None,
        order_class: str = None,
        take_profit: dict = None,
        stop_loss: dict = None,
        trail_price: str = None,
        trail_percent: str = None,
        on_behalf_of: str = None,
    ) -> Order:
        if on_behalf_of:
            return await self._order_on_behalf(
                symbol,
                qty,
                side,
                order_type,
                time_in_force,
                limit_price,
                stop_price,
                client_order_id,
                extended_hours,
                order_class,
                take_profit,
                stop_loss,
                trail_price,
                trail_percent,
                on_behalf_of,
            )
        else:
            return await self._personal_submit(
                symbol,
                qty,
                side,
                order_type,
                time_in_force,
                limit_price,
                stop_price,
                client_order_id,
                extended_hours,
                order_class,
                take_profit,
                stop_loss,
                trail_price,
                trail_percent,
                on_behalf_of,
            )

    @classmethod
    def _trade_from_dict(cls, trade_dict: Entity) -> Optional[Trade]:
        if trade_dict.event == "new":
            return None

        symbol = trade_dict.order["symbol"].lower().replace("/", "")
        return Trade(
            order_id=trade_dict.order["id"],
            symbol=symbol,
            event=Order.EventType.canceled
            if trade_dict.event
            in ["canceled", "suspended", "expired", "cancel_rejected"]
            else Order.EventType.rejected
            if trade_dict.event == "rejected"
            else Order.EventType.fill
            if trade_dict.event == "fill"
            else Order.EventType.partial_fill
            if trade_dict.event == "partial_fill"
            else Order.EventType.other,
            filled_qty=float(trade_dict.qty)
            if trade_dict.event in ["fill", "partial_fill"]
            else 0.0,
            trade_fee=0.0,
            filled_avg_price=float(
                trade_dict.order["filled_avg_price"] or 0.0
            ),
            liquidity="",
            updated_at=pd.Timestamp(
                ts_input=trade_dict.order["updated_at"],
                unit="ms",
                tz="US/Eastern",
            ),
            side=Order.FillSide[trade_dict.order["side"]],
        )

    @classmethod
    async def trade_update_handler(cls, data):
        try:
            # cls.get_instance().queues[symbol].put(
            #    data.__dict__["_raw"], timeout=1
            # )
            trade = cls._trade_from_dict(data)
            if not trade:
                return

            to_send = {
                "EV": "trade_update",
                "symbol": trade.symbol,
                "trade": trade.__dict__,
            }
            for q in cls.get_instance().queues.get_allqueues():
                q.put(to_send, timeout=1)

        except queue.Full as f:
            tlog(
                f"[EXCEPTION] process_message(): queue for {trade.symbol} is FULL:{f}, sleeping for 2 seconds and re-trying."
            )
            raise
        # except AssertionError:
        #    for q in cls.get_instance().queues.get_allqueues():
        #        q.put(data.__dict__["_raw"], timeout=1)
        except Exception as e:
            tlog(f"[EXCEPTION] process_message(): exception {e}")
            if config.debug_enabled:
                traceback.print_exc()



import json
import queue
import traceback
from datetime import date, datetime
from typing import Callable, Dict, List, Optional

import numpy as np
import pandas as pd
import requests
from polygon import RESTClient, WebSocketClient

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog
from liualgotrader.common.types import QueueMapper, TimeScale, WSEventType
from liualgotrader.data.data_base import DataAPI
from liualgotrader.data.streaming_base import StreamingAPI


class PolygonData(DataAPI):
    def __init__(self):
        self.polygon_rest_client = RESTClient(config.polygon_api_key)
        if not self.polygon_rest_client:
            raise AssertionError(
                "Failed to authenticate Polygon restful client"
            )

    async def get_market_snapshot(
        self, filter_func: Optional[Callable]
    ) -> List[Dict]:
        if not self.polygon_rest_client:
            raise AssertionError("Must call w/ authenticated polygon client")
        # this API endpoint requires at least starter subscriptions from Polygon
        data = self.polygon_rest_client.stocks_equities_snapshot_all_tickers()
        return (
            list(filter(filter_func, data.tickers))
            if filter_func is not None
            else data.tickers
        )

    def get_symbols(self) -> List[str]:
        if not self.polygon_rest_client:
            raise AssertionError("Must call w/ authenticated polygon client")
        # parse symbols on the first page
        data = self.polygon_rest_client.reference_tickers_v3(
            limit=1000, active=True
        )
        # use set to deduplicate in case paginated response return duplicate symbols
        symbols = {d["ticker"] for d in data.results}
        next_url = f"{data.next_url}&apiKey={config.polygon_api_key}"
        # parse the pagination
        while True:
            response = requests.get(next_url).json()
            if "next_url" not in response:
                break
            symbols.update([d["ticker"] for d in response["results"]])
            next_url = (
                f"{response['next_url']}&apiKey={config.polygon_api_key}"
            )
        return list(symbols)

    def get_symbol_data(
        self,
        symbol: str,
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> pd.DataFrame:
        if not self.polygon_rest_client:
            raise AssertionError("Must call w/ authenticated polygon client")

        data = self.polygon_rest_client.stocks_equities_aggregates(
            symbol, 1, scale.name, start, end, unadjusted=False, limit=50000
        )
        if not data or not hasattr(data, "results"):
            raise ValueError(
                f"[ERROR] {symbol} has no data for {start} to {end} w {scale.name}"
            )

        d = {
            pd.Timestamp(result["t"], unit="ms", tz="America/New_York"): [
                result.get("o"),
                result.get("h"),
                result.get("l"),
                result.get("c"),
                result.get("v"),
                result.get("vw"),
                result.get("n"),
            ]
            for result in data.results
        }
        df = pd.DataFrame.from_dict(
            d,
            orient="index",
            columns=[
                "open",
                "high",
                "low",
                "close",
                "volume",
                "average",
                "count",
            ],
        )
        df["vwap"] = np.NaN
        return df

    def get_symbols_data(
        self,
        symbols: List[str],
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> Dict[str, pd.DataFrame]:
        raise NotImplementedError("get_symbols_data")

    def get_last_trading(self, symbol: str) -> datetime:
        snapshot_data = (
            self.polygon_rest_client.stocks_equities_snapshot_single_ticker(
                symbol
            )
        )
        return pd.Timestamp(
            snapshot_data.ticker.last_trade.timestamp_of_this_trade,
            unit="ns",
            tz="America/New_York",
        )

    def get_trading_day(
        self, symbol: str, now: datetime, offset: int
    ) -> datetime:
        raise NotImplementedError("get_trading_day")

    def trading_days_slice(self, symbol: str, slice) -> slice:
        raise NotImplementedError("trading_days_slice")

    def num_trading_minutes(self, symbol: str, start: date, end: date) -> int:
        raise NotImplementedError("num_trading_minutes")

    def num_trading_days(self, symbol: str, start: date, end: date) -> int:
        raise NotImplementedError("num_trading_days")

    def get_max_data_points_per_load(self) -> int:
        raise NotImplementedError("get_max_data_points_per_load")


class PolygonStream(StreamingAPI):
    def __init__(self, queues: QueueMapper):
        self.polygon_ws_client = WebSocketClient(
            auth_key=config.polygon_api_key,
            process_message=PolygonStream.process_message,
            on_close=PolygonStream.on_close,
            on_error=PolygonStream.on_error,
        )
        if not self.polygon_ws_client:
            raise AssertionError(
                "Failed to authenticate Polygon web_socket client"
            )
        self.polygon_ws_client.run()
        super().__init__(queues)

    async def subscribe(
        self, symbols: List[str], events: List[WSEventType]
    ) -> bool:
        args = []
        for symbol in symbols:
            for event in events:
                if event == WSEventType.SEC_AGG:
                    action = "A"
                elif event == WSEventType.MIN_AGG:
                    action = "AM"
                elif event == WSEventType.TRADE:
                    action = "T"
                elif event == WSEventType.QUOTE:
                    action = "Q"

                args.append(f"{action}.{symbol}")

        tlog(f"subscribe(): adding subscription {args}")
        self.polygon_ws_client.subscribe(*args)

        return True

    async def run(self):
        pass

    async def unsubscribe(self, symbol: str) -> bool:
        raise NotImplementedError("not implemented yet")

    async def close(
        self,
    ) -> None:
        self.polygon_ws_client.close_connection()

    @classmethod
    def handle_event(cls, event: Dict):
        print("event:", event)
        try:
            event["EV"] = event["ev"]
            if "s" in event:
                event["start"] = event["s"]
            if "o" in event:
                event["open"] = event["o"]
            if "h" in event:
                event["high"] = event["h"]
            if "l" in event:
                event["low"] = event["l"]
            if "c" in event:
                event["close"] = event["c"]
            if "v" in event:
                event["volume"] = event["v"]
            if "vw" in event:
                event["vwap"] = event["vw"]
            if "a" in event:
                event["average"] = event["a"]
            if "av" in event:
                event["totalvolume"] = event["av"]
            if "sym" in event:
                event["symbol"] = event["sym"]
            if "z" in event:
                event["count"] = event["z"]
            cls.get_instance().queues[event["sym"]].put(event, timeout=1)
        except queue.Full as f:
            tlog(
                f"[EXCEPTION] process_message(): queue for {event['sym']} is FULL:{f}, sleeping for 2 seconds and re-trying."
            )
            raise
        except Exception as e:
            tlog(
                f"[EXCEPTION] process_message(): exception of type {type(e).__name__} with args {e.args}"
            )
            traceback.print_exc()

    @classmethod
    def process_message(cls, message):
        payload = json.loads(message)
        for event in payload:
            if event["ev"] in ("A", "AM", "T", "Q"):
                cls.handle_event(event)

    @classmethod
    def on_error(cls, ws, error):
        tlog(f"[ERROR] on_error(): {error}")

    @classmethod
    def on_close(cls, ws, close_status_code, close_msg):
        tlog(f"[INFO] on_close() called with {close_status_code, close_msg}")



"""Save trade details to repository"""
import json
from datetime import datetime
from typing import Dict, List, Tuple

from asyncpg.pool import Pool

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog


class NewTrade:
    def __init__(
        self,
        algo_run_id: int,
        symbol: str,
        operation: str,
        qty: float,
        price: float,
        indicators: dict,
    ):
        """
        create a new_trade object
        :param algo_run_id: id of the algorithm making the transaction
        :param symbol: stock symbol
        :param operation: buy or sell
        :param qty: amount being purchased
        :param price: buy price
        :param indicators: buy indicators
        """
        self.algo_run_id = algo_run_id
        self.symbol = symbol
        self.qty = qty
        self.price = price
        self.indicators = indicators
        self.operation = operation
        self.trade_id = None

    async def save(
        self,
        pool: Pool,
        client_buy_time: str,
        stop_price=None,
        target_price=None,
        trade_fee=0.0,
    ):
        async with pool.acquire() as con:
            async with con.transaction():
                try:
                    indicators_s = json.dumps(
                        self.indicators or {}, allow_nan=False
                    )
                except ValueError:
                    indicators_s = json.dumps({})

                self.trade_id = await con.fetchval(
                    """
                        INSERT INTO new_trades (algo_run_id, symbol, operation, qty, price, indicators, client_time, stop_price, target_price, trade_fee)
                        VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10)
                        RETURNING trade_id
                    """,
                    self.algo_run_id,
                    self.symbol,
                    self.operation,
                    self.qty,
                    self.price,
                    indicators_s,
                    client_buy_time,
                    stop_price,
                    target_price,
                    trade_fee,
                )

    @classmethod
    async def expire_trade(cls, pool: Pool, trade_id: int) -> None:
        async with pool.acquire() as con:
            async with con.transaction():
                await con.execute(
                    """
                        UPDATE new_trades SET expire_tstamp='now()' WHERE trade_id=$1
                    """,
                    trade_id,
                )

    @classmethod
    async def load_latest(
        cls, pool: Pool, symbol: str, strategy_name: str
    ) -> Tuple[int, float, float, float, Dict, datetime]:
        async with pool.acquire() as con:
            async with con.transaction():
                row = await con.fetchrow(
                    """
                        SELECT t.algo_run_id, t.price, t.stop_price, t.target_price, t.indicators, t.tstamp 
                        FROM new_trades as t, algo_run as a
                        WHERE 
                            t.algo_run_id=a.algo_run_id AND
                            a.algo_name=$2 AND
                            symbol=$1 
                        ORDER BY tstamp DESC LIMIT 1
                    """,
                    symbol,
                    strategy_name,
                )

                if row:
                    return (
                        int(row[0]),
                        float(row[1]),
                        float(row[2] or 0.0),
                        float(row[3] or 0.0),
                        json.loads(row[4] or "{}"),
                        row[5],
                    )
                tlog(f"{symbol} no data for strategy {strategy_name}")
                raise ValueError

    @classmethod
    async def get_run_symbols(
        cls, run_id: int, pool: Pool = None
    ) -> List[str]:
        rc: List = []
        if not pool:
            pool = config.db_conn_pool
        async with pool.acquire() as con:
            async with con.transaction():
                rows = await con.fetch(
                    """
                        SELECT DISTINCT symbol
                        FROM new_trades
                        WHERE algo_run_id = $1 and expire_tstamp is null
                    """,
                    run_id,
                )

                if rows:
                    rc = [row[0] for row in rows]

        return rc

    @classmethod
    async def rename_algo_run_id(
        cls, new_run_id: int, old_run_id: int, symbol: str, pool: Pool = None
    ) -> None:
        if not pool:
            pool = config.db_conn_pool

        async with pool.acquire() as con:
            async with con.transaction():
                await con.execute(
                    """
                        UPDATE 
                            new_trades 
                        SET 
                            algo_run_id=$1 
                        WHERE 
                            algo_run_id=$2 AND
                            symbol=$3
                    """,
                    new_run_id,
                    old_run_id,
                    symbol,
                )

    @classmethod
    async def get_latest_algo_run_id(
        cls, symbol: str, pool: Pool = None
    ) -> int:
        if not pool:
            pool = config.db_conn_pool
        async with pool.acquire() as con:
            async with con.transaction():
                val = await con.fetchval(
                    """
                        SELECT  algo_run_id
                        FROM new_trades
                        WHERE symbol = $1
                        ORDER BY tstamp DESC
                        LIMIT 1
                    """,
                    symbol,
                )

                return val



import json
from dataclasses import dataclass
from datetime import date, datetime
from typing import Dict, List, Optional

from asyncpg.exceptions import TooManyConnectionsError
from asyncpg.pool import Pool

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog


@dataclass
class TickerData:
    name: str
    symbol: str
    description: str
    tags: List[str]
    similar_tickers: List[str]
    industry: str
    sector: str
    exchange: str

    @classmethod
    async def load_symbols(cls, pool: Pool = None) -> List[str]:
        if not pool:
            pool = config.db_conn_pool

        async with pool.acquire() as con:
            async with con.transaction():
                rows = await con.fetch(
                    """
                        SELECT symbol
                        FROM ticker_data 
                    """,
                )

                if rows:
                    return [row[0] for row in rows]
                else:
                    raise Exception("no data")

    async def save(self, pool: Pool) -> bool:
        try:
            async with pool.acquire() as con:
                async with con.transaction():
                    await con.execute(
                        """
                            INSERT INTO ticker_data (symbol, name, description, tags,
                                                     similar_tickers, industry, sector, exchange)
                            VALUES ($1, $2, $3, $4, $5, $6, $7, $8)
                            ON CONFLICT (symbol)
                            DO UPDATE
                                SET name=$2, description=$3, tags=$4, similar_tickers=$5, industry=$6,
                                    sector=$7, exchange=$8, modify_tstamp='now()'
                        """,
                        self.symbol,
                        self.name,
                        self.description,
                        self.tags,
                        self.similar_tickers,
                        self.industry,
                        self.sector,
                        self.exchange,
                    )
            return True
        except TooManyConnectionsError as e:
            tlog(
                f"too many db connections: {e}. Failed to write ticker {self.symbol}, will re-try"
            )

        return False


@dataclass
class StockOhlc:
    symbol: str
    symbol_date: date
    open: float
    high: float
    low: float
    close: float
    volume: int
    indicators: Dict
    create_tstamp: Optional[datetime] = None

    @classmethod
    async def check_stock_date_exists(
        cls, symbol: str, symbol_date: date, pool: Pool = None
    ) -> bool:
        if not pool:
            pool = config.db_conn_pool

        async with pool.acquire() as con:
            async with con.transaction():
                val = await con.fetchval(
                    """
                        SELECT 
                            COUNT(symbol)
                        FROM 
                            stock_ohlc
                        WHERE 
                            symbol_date = $1 AND
                            symbol = $2
                        LIMIT 1
                    """,
                    symbol_date,
                    symbol,
                )
                return val > 0

    @classmethod
    async def get_latest_date(cls, symbol: str, pool: Pool = None) -> date:
        if not pool:
            pool = config.db_conn_pool

        async with pool.acquire() as con:
            async with con.transaction():
                val = await con.fetchval(
                    """
                        SELECT symbol_date
                        FROM stock_ohlc
                        WHERE symbol=$1 
                        ORDER by symbol_date DESC
                        LIMIT 1
                    """,
                    symbol,
                )
                return val

    @classmethod
    async def load_by_date(
        cls, symbol_date: date, pool: Pool = None
    ) -> Dict[str, object]:
        if not pool:
            pool = config.db_conn_pool

        async with pool.acquire() as con:
            async with con.transaction():
                rows = await con.fetch(
                    """
                        SELECT symbol, symbol_date, open, high, low, close, volume, indicators
                        FROM stock_ohlc
                        WHERE symbol_date = $1 
                    """,
                    symbol_date,
                )

                rc: Dict[str, object] = {}

                for x in rows:
                    rc[x[0]] = StockOhlc(
                        symbol=x[0],
                        symbol_date=x[1],
                        open=x[2],
                        high=x[3],
                        low=x[4],
                        close=x[5],
                        volume=x[6],
                        indicators=json.loads(x[7]),
                    )

                return rc

    async def save(
        self,
        pool: Pool = None,
    ) -> None:
        if not pool:
            pool = config.db_conn_pool

        async with pool.acquire() as con:
            async with con.transaction():
                await con.execute(
                    """
                        INSERT INTO stock_ohlc (symbol, symbol_date, open, high, low, close, volume, indicators)
                        VALUES ($1, $2, $3, $4, $5, $6, $7, $8)
                        ON CONFLICT (symbol, symbol_date)
                        DO UPDATE
                            SET open=$3, high=$4, low=$5, close=$6, volume=$7, indicators=$8, modify_tstamp='now()'
                    """,
                    self.symbol,
                    self.symbol_date,
                    self.open,
                    self.high,
                    self.low,
                    self.close,
                    self.volume,
                    json.dumps(self.indicators),
                )



from typing import Dict

from liualgotrader.common.types import AssetType

assets_details: Dict[str, Dict] = {
    "btc/usd": {
        "type": AssetType.CRYPTO,
        "min_order_size": 0.00001,
        "tick_precision": 8,
    },
    "btcusd": {
        "type": AssetType.CRYPTO,
        "min_order_size": 0.00001,
        "tick_precision": 8,
    },
    "eth/usd": {
        "type": AssetType.CRYPTO,
        "min_order_size": 0.001,
        "tick_precision": 6,
    },
    "ethusd": {
        "type": AssetType.CRYPTO,
        "min_order_size": 0.001,
        "tick_precision": 6,
    },
}


def get_asset_precision(asset_name: str) -> int:
    asset_name = asset_name.lower()
    if asset_name not in assets_details:
        raise ValueError(f"asset name {asset_name} is undefined")

    return assets_details[asset_name]["tick_precision"]


def round_asset(asset_name: str, value: float) -> float:
    asset_name = asset_name.lower()
    return round(value, get_asset_precision(asset_name))


def get_asset_min_qty(asset_name: str) -> float:
    asset_name = asset_name.lower()
    if asset_name not in assets_details:
        raise ValueError(f"asset name {asset_name} is undefined")

    return assets_details[asset_name]["min_order_size"]



import asyncio
import inspect
import traceback
import uuid
from datetime import date, datetime, timedelta
from typing import Dict, List, Optional

import alpaca_trade_api as tradeapi
import pandas as pd
from pytz import timezone

from liualgotrader.common import config, trading_data
from liualgotrader.common.data_loader import DataLoader  # type: ignore
from liualgotrader.common.database import create_db_connection
from liualgotrader.common.tlog import tlog, tlog_exception
from liualgotrader.common.types import AssetType, TimeScale
from liualgotrader.models.new_trades import NewTrade
from liualgotrader.scanners.base import Scanner  # type: ignore
from liualgotrader.scanners_runner import create_momentum_scanner
from liualgotrader.strategies.base import Strategy
from liualgotrader.trading.base import Trader
from liualgotrader.trading.trader_factory import trader_factory

run_scanners: Dict[Scanner, datetime] = {}
symbol_data: Dict[str, pd.DataFrame] = {}
portfolio_value: float = 100000


async def create_scanners(
    data_loader: DataLoader,
    scanners_conf: Dict,
    scanner_names: Optional[List],
) -> List[Scanner]:
    scanners: List = []
    for scanner_name in scanners_conf:
        if scanner_names and scanner_name not in scanner_names:
            continue
        tlog(f"scanner {scanner_name} selected")
        if scanner_name == "momentum":
            scanners.append(
                await create_momentum_scanner(
                    None, data_loader, scanners_conf[scanner_name]  # type: ignore
                )
            )
        else:
            scanners.append(
                await Scanner.get_scanner(
                    data_loader=data_loader,
                    scanner_name=scanner_name,
                    scanner_details=scanners_conf[scanner_name],
                )
            )

    return scanners


async def create_strategies(
    uid: str,
    conf_dict: Dict,
    dl: DataLoader,
    strategy_names: Optional[List],
) -> List[Strategy]:
    strategies = []
    config.env = "BACKTEST"

    for strategy_name in conf_dict:
        if strategy_names and strategy_name not in strategy_names:
            continue
        strategy_details = conf_dict[strategy_name]
        strategies.append(
            await Strategy.get_strategy(
                batch_id=uid,
                strategy_name=strategy_name,
                strategy_details=strategy_details,
                data_loader=dl,
            )
        )

    return strategies


async def do_scanners(
    now: datetime, scanners: List[Scanner], symbols: Dict
) -> Dict:
    for scanner in scanners:
        if scanner in run_scanners:
            if not scanner.recurrence:
                if now.date() == run_scanners[scanner].date():
                    continue
            elif (now - run_scanners[scanner]) < scanner.recurrence:
                continue

        run_scanners[scanner] = now
        new_symbols = await scanner.run(back_time=now)
        target_strategy_name = scanner.target_strategy_name

        target_strategy_name = target_strategy_name or "_all"

        symbols[target_strategy_name] = new_symbols

        # list(
        #    set(symbols.get(target_strategy_name, [])).union(set(new_symbols))
        # )

    return symbols


async def calculate_execution_price(
    symbol: str, data_loader: DataLoader, what: Dict, now: datetime
) -> float:
    if what["type"] == "market":
        return data_loader[symbol].close[now]

    price_limit = float(what["limit_price"])
    if what["side"] == "buy":
        if data_loader[symbol].close[now] <= price_limit:
            return data_loader[symbol].close[now]
        else:
            raise Exception(
                f"can not buy: limit price {price_limit} below market price {data_loader[symbol].close[now]}"
            )

    if data_loader[symbol].close[now] >= price_limit:
        return price_limit
    else:
        raise Exception(
            f"can not sell: limit price {price_limit} above market price {data_loader[symbol].close[now]}"
        )


async def do_strategy_result(
    data_loader: DataLoader,
    strategy: Strategy,
    symbol: str,
    now: datetime,
    what: Dict,
    buy_fee_percentage: float = 0.0,
    sell_fee_percentage: float = 0.0,
) -> bool:
    global portfolio_value

    qty = float(what["qty"])
    sign = (
        1
        if (
            what["side"] == "buy"
            and qty > 0
            or what["side"] == "sell"
            and qty < 0
        )
        else -1
    )

    try:
        price = await calculate_execution_price(
            symbol=symbol, data_loader=data_loader, what=what, now=now
        )
        if what["side"] == "buy":
            fee = price * qty * buy_fee_percentage / 100.0

            sig = inspect.signature(strategy.buy_callback)
            if "trade_fee" in sig.parameters:
                await strategy.buy_callback(symbol, price, qty, now, fee)
            else:
                await strategy.buy_callback(symbol, price, qty, now)

        elif what["side"] == "sell":
            fee = price * qty * sell_fee_percentage / 100.0
            sig = inspect.signature(strategy.buy_callback)
            if "trade_fee" in sig.parameters:
                await strategy.sell_callback(symbol, price, qty, now, fee)
            else:
                await strategy.sell_callback(symbol, price, qty, now)
    except Exception as e:
        tlog(
            f"do_strategy_result({symbol}, {what}, {now}) failed w/ {e}. operation not executed"
        )
        return False

    try:
        trading_data.positions[symbol] = (
            trading_data.positions[symbol] + sign * qty
        )
    except KeyError:
        trading_data.positions[symbol] = sign * qty

    trading_data.buy_time[symbol] = now.replace(second=0, microsecond=0)
    trading_data.last_used_strategy[symbol] = strategy

    db_trade = NewTrade(
        algo_run_id=strategy.algo_run.run_id,
        symbol=symbol,
        qty=qty,
        operation=what["side"],
        price=price,
        indicators={},
    )

    await db_trade.save(
        config.db_conn_pool,
        str(now),
        trading_data.stop_prices[symbol]
        if symbol in trading_data.stop_prices
        else 0.0,
        trading_data.target_prices[symbol]
        if symbol in trading_data.target_prices
        else 0.0,
        fee,
    )

    return True


async def do_strategy_all(
    data_loader: DataLoader,
    now: pd.Timestamp,
    strategy: Strategy,
    symbols: List[str],
    trader: Trader,
    buy_fee_percentage: float,
    sell_fee_percentage: float,
):
    try:
        sig = inspect.signature(strategy.run_all)
        param = {
            "symbols_position": dict(
                {symbol: 0 for symbol in symbols}, **trading_data.positions
            ),
            "now": now.to_pydatetime(),
            "portfolio_value": portfolio_value,
            "backtesting": True,
            "data_loader": data_loader,
            "trader": trader,
        }
        if "fee_buy_percentage" in sig.parameters:
            param["fee_buy_percentage"] = buy_fee_percentage
        if "fee_sell_percentage" in sig.parameters:
            param["fee_sell_percentage"] = sell_fee_percentage
        do = await strategy.run_all(**param)
        items = list(do.items())
        items.sort(key=lambda x: int(x[1]["side"] == "buy"))
        for symbol, what in items:
            await do_strategy_result(
                data_loader,
                strategy,
                symbol,
                now,
                what,
                buy_fee_percentage,
                sell_fee_percentage,
            )

    except Exception as e:
        tlog(f"[Exception] {now} {strategy}->{e}")
        traceback.print_exc()
        raise


async def do_strategy_by_symbol(
    data_loader: DataLoader,
    now: pd.Timestamp,
    strategy: Strategy,
    symbols: List[str],
):
    global portfolio_value
    for symbol in symbols:
        try:
            _ = data_loader[symbol][now - timedelta(days=1) : now]  # type: ignore
            do, what = await strategy.run(
                symbol=symbol,
                shortable=True,
                position=trading_data.positions.get(symbol, 0.0),
                now=now,
                portfolio_value=portfolio_value,
                backtesting=True,
                minute_history=data_loader[symbol].symbol_data[:now],  # type: ignore
            )

            if do:
                await do_strategy_result(
                    data_loader, strategy, symbol, now, what
                )
        except ValueError:
            tlog(
                f"symbol {symbol} for now have data in the range {now - timedelta(days=1)}-{now}"
            )
        except Exception as e:
            tlog(f"[Exception] {now} {strategy}({symbol}):`{e}`")
            if config.debug_enabled:
                traceback.print_exc()
            raise


async def do_strategy(
    data_loader: DataLoader,
    now: pd.Timestamp,
    strategy: Strategy,
    symbols: List[str],
    trader: Trader,
    buy_fee_percentage: float,
    sell_fee_percentage: float,
):
    if await strategy.should_run_all():
        await do_strategy_all(
            data_loader,
            now,
            strategy,
            symbols,
            trader,
            buy_fee_percentage,
            sell_fee_percentage,
        )
    else:
        await do_strategy_by_symbol(data_loader, now, strategy, symbols)


def get_day_start_end(asset_type, day):
    if asset_type == AssetType.US_EQUITIES:
        day_start = day.date.replace(
            hour=day.open.hour,
            minute=day.open.minute,
            tzinfo=timezone("America/New_York"),
        )
        day_end = day.date.replace(
            hour=day.close.hour,
            minute=day.close.minute,
            tzinfo=timezone("America/New_York"),
        )
    elif asset_type == AssetType.CRYPTO:
        day_start = pd.Timestamp(
            datetime.combine(day, datetime.min.time()),
            tzinfo=timezone("America/New_York"),
        )
        day_end = pd.Timestamp(
            datetime.combine(day, datetime.max.time()),
            tzinfo=timezone("America/New_York"),
        )
    else:
        raise AssertionError(
            f"get_day_start_end(): asset type {asset_type} not yet supported"
        )

    return day_start, day_end


async def backtest_day(
    day,
    scanners,
    symbols,
    strategies,
    scale,
    data_loader,
    asset_type: AssetType,
    trader: Trader,
    buy_fee_percentage: float,
    sell_fee_percentage: float,
):
    day_start, day_end = get_day_start_end(asset_type, day)
    print(day_start, day_end)
    current_time = day_start
    prefetched: List[str] = []
    while current_time < day_end:
        symbols = await do_scanners(current_time, scanners, symbols)
        trading_data.positions = {
            symbol: trading_data.positions[symbol]
            for symbol in trading_data.positions
            if trading_data.positions[symbol] != 0
        }

        prefetch = [
            item
            for sublist in symbols.values()
            for item in sublist
            if item not in prefetched
        ]
        for symbol in prefetch:
            tlog(f"Prefetch data for {symbol}@{day_start}-{day_end}")

            data_loader[symbol][day_start:day_end]
            prefetched.append(symbol)

        for strategy in strategies:
            try:
                trading_data.positions |= {
                    symbol: 0
                    for symbol in symbols.get(strategy.name, [])
                    + symbols.get("_all", [])
                    if symbol not in trading_data.positions
                }

            except TypeError as e:
                if config.debug_enabled:
                    tlog_exception("backtest_day")

                tlog(
                    f"[EXCEPTION] {e} (hint: check scanner(s) return list of symbols?)"
                )
                raise

            strategy_symbols = list(
                set(symbols.get("_all", [])).union(
                    set(symbols.get(strategy.name, []))
                )
            )
            await do_strategy(
                data_loader,
                current_time,
                strategy,
                strategy_symbols,
                trader,
                buy_fee_percentage,
                sell_fee_percentage,
            )

        current_time += timedelta(seconds=scale.value)


async def backtest_time_range(
    from_date: date,
    to_date: date,
    scale: TimeScale,
    data_loader: DataLoader,
    buy_fee_percentage: float,
    sell_fee_percentage: float,
    asset_type: AssetType,
    scanners: Optional[List] = None,
    strategies: Optional[List] = None,
):
    trade_api = tradeapi.REST(
        key_id=config.alpaca_api_key, secret_key=config.alpaca_api_secret
    )
    if asset_type == AssetType.US_EQUITIES:
        calendars = trade_api.get_calendar(str(from_date), str(to_date))
    elif asset_type == AssetType.CRYPTO:
        calendars = [
            t.date() for t in pd.date_range(from_date, to_date).to_list()
        ]
    else:
        raise AssertionError(f"Asset type {asset_type} not supported yet")

    symbols: Dict = {}

    for day in calendars:
        await backtest_day(
            day=day,
            scanners=scanners,
            symbols=symbols,
            strategies=strategies,
            scale=scale,
            data_loader=data_loader,
            asset_type=asset_type,
            trader=trader_factory(),
            buy_fee_percentage=buy_fee_percentage,
            sell_fee_percentage=sell_fee_percentage,
        )


async def backtest_main(
    uid: str,
    from_date: date,
    to_date: date,
    scale: TimeScale,
    tradeplan: Dict,
    buy_fee_percentage: float,
    sell_fee_percentage: float,
    asset_type: AssetType,
    scanners: Optional[List] = None,
    strategies: Optional[List] = None,
) -> None:
    tlog(
        f"Starting back-test from {from_date} to {to_date} with time scale {scale}"
    )
    if scanners:
        tlog(f"with scanners:{scanners}")
    if strategies:
        tlog(f"with strategies:{strategies}")

    if "portfolio_value" in tradeplan:
        global portfolio_value
        portfolio_value = tradeplan["portfolio_value"]

    await create_db_connection()

    data_loader = DataLoader(scale)

    scanners = (
        await create_scanners(data_loader, tradeplan["scanners"], scanners)
        if "scanners" in tradeplan
        else []
    )
    tlog(f"instantiated {len(scanners)} scanners")
    strategies = await create_strategies(
        uid, tradeplan["strategies"], data_loader, strategies
    )
    tlog(f"instantiated {len(strategies)} strategies")

    await backtest_time_range(
        from_date=from_date,
        to_date=to_date,
        scale=scale,
        data_loader=data_loader,
        buy_fee_percentage=buy_fee_percentage,
        sell_fee_percentage=sell_fee_percentage,
        asset_type=asset_type,
        scanners=scanners,
        strategies=strategies,
    )


def backtest(
    from_date: date,
    to_date: date,
    scale: TimeScale,
    config: Dict,
    scanners: Optional[List],
    strategies: Optional[List],
    asset_type: AssetType,
    buy_fee_percentage: float,
    sell_fee_percentage: float,
) -> str:
    uid = str(uuid.uuid4())
    try:
        if not asyncio.get_event_loop().is_closed():
            asyncio.get_event_loop().close()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(asyncio.new_event_loop())
        loop.run_until_complete(
            backtest_main(
                uid,
                from_date,
                to_date,
                scale,
                config,
                buy_fee_percentage,
                sell_fee_percentage,
                asset_type,
                scanners,
                strategies,
            )
        )
    except KeyboardInterrupt:
        tlog("backtest() - Caught KeyboardInterrupt")
    except Exception as e:
        tlog(
            f"backtest() - exception of type {type(e).__name__} with args {e.args}"
        )
        traceback.print_exc()
    finally:
        print("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=")
        print(f"new batch-id: {uid}")

        return uid



from datetime import datetime, timedelta
from enum import Enum
from typing import List, Optional

import numpy as np
import pandas as pd
import pytz
from pandas import DataFrame as df
from pandas import Timestamp as ts

from liualgotrader.common import config

est = pytz.timezone("US/Eastern")


class StopRangeType(Enum):
    LAST_100_MINUTES = 1
    LAST_2_HOURS = 2
    LAST_3_HOURS = 3
    DAILY = 10
    WEEKLY = 20
    DATE_RANGE = 50


def grouper(iterable):
    prev = None
    group = []
    for item in iterable:

        if (
            not prev
            or -config.group_margin
            <= float(item - prev) / prev
            <= config.group_margin
        ):
            group.append(item)
        else:
            yield group
            group = [item]
        prev = item
    if group:
        yield group


async def find_resistances(
    symbol: str,
    strategy_name: str,
    current_value: float,
    minute_history: df,
    debug=False,
) -> Optional[List[float]]:
    """calculate supports"""

    est = pytz.timezone("America/New_York")
    back_time = ts(datetime.now(est)).to_pydatetime() - timedelta(days=3)
    back_time_index = minute_history["close"].index.get_loc(
        back_time, method="nearest"
    )

    series = (
        minute_history["close"][back_time_index:]
        .dropna()
        .between_time("9:30", "16:00")
        .resample("15min")
        .max()
    ).dropna()

    diff = np.diff(series.values)
    high_index = np.where((diff[:-1] >= 0) & (diff[1:] <= 0))[0] + 1
    if len(high_index) > 0:
        local_maximas = sorted(
            [series[i] for i in high_index if series[i] >= current_value]
        )
        if len(local_maximas) > 0:
            return local_maximas

    return None


def find_supports(
    current_value,
    minute_history,
    now: datetime,
    range_type: StopRangeType = StopRangeType.LAST_100_MINUTES,
):
    # get low Series based on select time-range
    if range_type == StopRangeType.DAILY:
        series = (
            minute_history["low"][
                ts(
                    now.replace(
                        hour=9, minute=30, second=0, microsecond=0, tzinfo=est
                    )
                ) :
            ]
            .resample("5min")
            .min()
        )
    elif range_type == StopRangeType.LAST_100_MINUTES:
        series = minute_history["low"][-100:].dropna().resample("5min").min()
        series = series[ts(now).floor("1D") :]
    elif range_type == StopRangeType.LAST_2_HOURS:
        series = minute_history["low"][-120:].dropna().resample("5min").min()
        series = series[ts(now).floor("1D") :]
    elif range_type == StopRangeType.LAST_3_HOURS:
        series = minute_history["low"][-180:].dropna().resample("5min").min()
        series = series[ts(now).floor("1D") :]
    else:
        raise NotImplementedError(
            f"stop-range type {range_type} is not implemented"
        )

    # find local minima
    diff = np.diff(series.values)
    low_index = np.where((diff[:-1] <= 0) & (diff[1:] > 0))[0] + 1
    if len(low_index) > 0:
        return [series[x] for x in low_index if series[x] < current_value]
    return None


def find_stop(
    current_value,
    minute_history,
    now: datetime,
    range_type: StopRangeType = StopRangeType.LAST_100_MINUTES,
):
    if range_type in (StopRangeType.DATE_RANGE, StopRangeType.WEEKLY):
        raise NotImplementedError(
            f"stop-range type {range_type} is not implemented"
        )

    if range_type == StopRangeType.DAILY:
        series = (
            minute_history["low"][
                ts(
                    now.replace(
                        hour=9, minute=30, second=0, microsecond=0, tzinfo=est
                    )
                ) :
            ]
            .dropna()
            .resample("5min")
            .min()
        )
    else:
        series = minute_history["low"][-100:].dropna().resample("5min").min()
        series = series[ts(now).floor("1D") :]

    diff = np.diff(series.values)
    low_index = np.where((diff[:-1] <= 0) & (diff[1:] > 0))[0] + 1
    if len(low_index) > 0:
        return series[low_index[-1]]  # - max(0.05, current_value * 0.02)
    return None  # current_value * config.default_stop


def get_local_maxima(
    series: pd.Series,
    debug=False,
) -> pd.Series:
    """calculate local maximal point"""
    if series.empty:
        return pd.Series([], dtype=np.float64)

    series = series.resample("5min").max()
    diff = np.diff(series.values)
    high_index = np.where((diff[:-1] >= 0) & (diff[1:] <= 0))[0] + 1

    return (
        pd.Series(
            index=[series.index[i] for i in high_index],
            data=[series[i] for i in high_index],
            dtype=np.float64,
        )
        if len(high_index) > 0
        else pd.Series([], dtype=np.float64)
    )



from enum import Enum

import pandas as pd


class ResampleRangeType(Enum):
    min_1 = 0
    min_2 = 1
    min_5 = 2
    min_10 = 3
    min_15 = 4


def resample(
    ohlc: pd.DataFrame, resample_range: ResampleRangeType
) -> pd.DataFrame:
    if ohlc.empty:
        return ohlc
    if resample_range == resample_range.min_1:
        return ohlc
    elif resample_range == resample_range.min_2:
        resample_str = "2min"
    elif resample_range == resample_range.min_5:
        resample_str = "5min"
    elif resample_range == resample_range.min_10:
        resample_str = "10min"
    elif resample_range == resample_range.min_15:
        resample_str = "15min"
    else:
        raise NotImplementedError(
            f"range type {resample_range} is not implemented"
        )

    close = ohlc.close.resample(resample_str).last()
    open = ohlc.open.resample(resample_str).first()
    high = ohlc.high.resample(resample_str).max()
    low = ohlc.low.resample(resample_str).min()
    volume = ohlc.volume.resample(resample_str).sum()

    return pd.concat(
        [
            open.rename("open"),
            high.rename("high"),
            low.rename("low"),
            close.rename("close"),
            volume.rename("volume"),
        ],
        axis=1,
    )
